"""
all the table models are added here for p8 schema and can be iterated and used to generate install scripts
we add system fields in the sql model such as timestamps and user ids
"""

import uuid
from pydantic import Field, model_validator
from pydantic.fields import FieldInfo
import typing
from ..AbstractModel import AbstractModel, BaseModel, AbstractEntityModel
from .. import inspection
from percolate.utils import make_uuid
import datetime
from .. import DefaultEmbeddingField, KeyField
from percolate.utils.names import EmbeddingProviders
import percolate as p8
import json
from percolate.utils import get_iso_timestamp
import hashlib
from enum import Enum
#we shouldnt really need a logger her but we have temp functions being exported to apis
from percolate.utils import logger
class Function(AbstractEntityModel):
    """Functions are external tools that agents can use. See field comments for context.
    Functions can be searched and used as LLM tools. 
    The function spec is derived from OpenAPI but adapted to the conventional used in LLMs
    """
         
    id: uuid.UUID | str = Field(description="A unique id in this case generated by the proxy and function name") 
    key: typing.Optional[str] = Field(description='optional key e.g operation id')
    name: str = Field(description="A friendly name that is unique within the proxy scope e.g. a single api or python library")
    verb:  typing.Optional[str] = Field(None,description= "The verb e.g. get, post etc")
    endpoint: typing.Optional[str] = Field(None, description="A callable endpoint in the case of REST")
    description: str = DefaultEmbeddingField('', description="A detailed description of the function - may be more comprehensive than the one within the function spec - this is semantically searchable")
    function_spec: dict = Field(description="A function description that is OpenAI and based on the OpenAPI spec")
    proxy_uri: str = Field(description='a reference to an api or library namespace that qualifies the named function')
    
    def __call__(self, *args, **kwargs):
        """
        Convenience to call any function via its proxy
        """
        return p8.get_proxy(self.proxy_uri).invoke(self, **kwargs)
        
    @classmethod
    def from_entity(cls, model:BaseException):
        """The function produces a callable agent stub that can be searched.
           In the database we use a native generic function handler and in python we load the agent from the name to call the agent.
           Agents are discoverable in this way and we can transfer context to other agents.
        """
        M:AbstractModel = AbstractModel.Abstracted(model)
    
        description = f"""
        ## Agent: {M.get_model_full_name()}
        ## Description
        ```
        {M.get_model_description()}
        ```
        ## Functions
        ```json
        {M.get_model_functions()}
        ```
        """
        
        """this must be the same interface as the Agent interface i.e. we relay to the agent.run(question='')"""
        def run(question:str):
            """
            send any query or request to the agent to execute the agent
            
            Args:
                question: a request to the agent
            """
            pass
        
        key = f"{M.get_model_full_name()}.run".replace('.','_')
        
        """the alias is important because we need to qualify functions used by agents and the proxy above is just a signature hint"""
        spec = cls.from_callable(run,alias=key).function_spec
        """names should not have '.'"""
        
        """for the name we qualify with run agent so its clear what we are doing"""
        return Function(name=key, 
                        key=key,
                        description=description,
                        function_spec=spec,
                        verb='get',
                        proxy_uri=f'p8agent/{M.get_model_full_name()}', #handler can use this - a p8_agent/agent.name - see interface proxy manager
                        endpoint='run')
        
    @classmethod
    def from_callable(cls, fn:typing.Callable, remove_untyped:bool=True, proxy_uri:str=None,alias:str=None):
        """construct function from callable - supply an alias if the function is not full qualified or if we need extra context"""
        def process_properties(properties: dict):
             
            untyped = []
            """google at least wont like some of these but they are redundant anyway"""
            for key, details in properties.items():
                for remove_field in ['title', 'default']:
                    if remove_field in details:
                        details.pop(remove_field)
       
                if 'anyOf' in details:
                    new_list = [t for t in details['anyOf'] if t['type']!= 'null']
                   
                    #temp ignore anything after the first type
                    if len(new_list) >= 1:
                        details['type'] = new_list[0]['type']
                        details.pop('anyOf')
                    else:
                        details.pop('anyOf')
                        details['oneOf'] = new_list
               
                if 'properties' in details:
                    process_properties(details['properties'])
                
                """for language models there is no point sending untyped information
                TODO: handle this better e.g. typed kwargs - specifying additional properties true might be enough but its unclear how to handle them anyway
                """
                if not details and remove_untyped:
                    untyped.append(key)
                

            
        def _map(f):
            """make sure the structure from pydantic is the same as used elsewhere for functions"""
            p = dict(f)
            if 'properties' in p:
                process_properties(p['properties'])
    
            name = p.pop('title')
            desc = p.pop('description') if 'description' in p else 'NO DESC'
            return {
                'name': alias or name,
                'parameters' :p,
                'description': desc
            }
        
        """we can pass this in e.g. for native. Lib is used for lib runtime or APIs for REST"""
        if not proxy_uri:
            proxy_uri = fn.__self__ if hasattr(fn, '__self__') else 'lib'
            if not isinstance(proxy_uri,str):
                proxy_uri = inspection.get_object_id(proxy_uri)
            
        s = _map(AbstractModel.create_model_from_function(fn).model_json_schema())
        key = s['name'] if not proxy_uri else f"{proxy_uri}.{s['name']}"
        id_md5_uuid =  uuid.uuid3(uuid.NAMESPACE_DNS, key)   
        return cls(id=str(id_md5_uuid), 
                   name = s['name'],
                   key  = key, 
                   endpoint=s['name'],
                   verb='get',
                   proxy_uri=proxy_uri,
                   spec=s, 
                   function_spec=s,
                   description=s['description'],
                   #this is ignored by Function but the runtime function can carry it
                   fn=fn)
        
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({ 'key': values['name'], 'proxy_uri':values['proxy_uri']})
        return values
        
   
class ApiProxy(AbstractEntityModel):
    """A list of proxies or APIs that have attached functions or endpoints - links to proxy_uri on the Function""" 
    id: typing.Optional[uuid.UUID | str] = Field(None, description="Will default to a hash of the uri")
    name: typing.Optional[str] = Field(None, description="A unique api friendly name")
    proxy_uri: str = Field(description='a reference to an api or library namespace that qualifies the named function')
    token: typing.Optional[str] = Field(None, description="the token to save")
    
            
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('name'):
            values['name'] = values['proxy_uri']
        if not values.get('id'):
            values['id'] = make_uuid(values['proxy_uri'])
        return values
        
    
    
class LanguageModelApi(AbstractEntityModel):
    """The Language model REST Apis are stored with tokens and scheme information.
    """
    
    id: uuid.UUID | str
    name: str = Field(description="A unique name for the model api e.g cerebras-llama3.1-8b")
    model: typing.Optional[str] = Field(None, description="The model name defaults to the name as they are often the same. the name can be unique based on a provider qualfier")
    scheme: typing.Optional[str] = Field('openai', description="In practice most LLM APIs use an openai scheme - currently `anthropic` and `google` can differ")
    completions_uri: str = Field(description="The api used for completions in chat and function calling. There may be other uris for other contexts")
    token_env_key: typing.Optional[str] = Field(None, description="Conventions are used to resolve keys from env or other services. Provide an alternative key")
    token: typing.Optional[str] = Field(None, description="It is not recommended to add tokens directly to the data but for convenience you might want to")
    
        
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('model'):
            values['model'] = values['name']
        return values
        
    
class Agent(AbstractEntityModel):
    """The agent model is a meta data object to persist agent metadata for search etc"""

    id: uuid.UUID| str  
    name: str
    category: typing.Optional[str] = Field(None,description="Simple property to filter agents by categories")
    description: str = DefaultEmbeddingField(description="The system prompt as markdown")
    spec: dict = Field(description="The model json schema")
    functions: typing.Optional[dict] = Field(description="The function that agent can call",default_factory=dict)
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        """we take these from the class and save them"""
        if not values.get('functions') and hasattr(cls, 'get_model_functions'):
            values['functions'] = cls.get_model_functions()
        return values
    
    
    def from_abstract_model(cls: BaseModel):
        """Given any pydantic model that behaves like an AbstractModel, get the meta object i.e. Agent"""
        cls:AbstractModel = AbstractModel.Abstracted(cls)
        name = cls.get_model_full_name()
        return Agent(id=make_uuid(name), 
                     name=name, 
                     functions=cls.get_model_functions(),
                     spec = cls.model_json_schema(),
                     description=cls.get_model_description())

class ModelField(AbstractEntityModel):
    """Fields are each field in any saved model/agent. 
    Fields are useful for describing system info such as for embeddings or for promoting.
    """
            
    id: typing.Optional[uuid.UUID|str] = Field(description="a unique key for the field e.g. field and entity key hashed")
    name: str = Field(description="The field name")
    entity_name: str
    field_type: str
    embedding_provider: typing.Optional[str] = Field(None, description="The embedding could be a multiple in future")
    description: typing.Optional[str] = None
    is_key: typing.Optional[bool] = Field(default=False, description="Indicate that the field is the primary key - our convention is the id field should be the primary key and be uuid and we use this to join embeddings")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        """this is a convenience for 'dont care' cases where you want an embedding but dont know the model name"""
        if values.get('embedding_provider') == 'default':
            values['embedding_provider'] = EmbeddingProviders.embedding_text_embedding_ada_002
        if not values.get('id'):
            values['id'] = make_uuid({'name': values['name'], 'entity_name': values['entity_name']})
        return values
    
    @classmethod
    def from_field_info(cls, key:str, field_type: FieldInfo, parent_model:BaseModel):
        """five a FieldInfo object map the field"""
        
        parent_model:AbstractModel = AbstractModel.Abstracted(parent_model)
        json_args = field_type.json_schema_extra or {}
        arg = inspection.get_innermost_args(field_type.annotation)
        arg = str(arg.__name__) if hasattr(arg, '__name__') else str(arg)
        return ModelField(name=key, 
                           entity_name= parent_model.get_model_full_name(), 
                           field_type= arg,
                           description=field_type.description,
                           **json_args)
        
    @staticmethod
    def get_fields_from_model(cls:BaseModel):
        """for any pydantic model generate the field collection"""
        cls = AbstractModel.Abstracted(cls)
        return [ModelField.from_field_info(key, value,parent_model=cls) for key, value in cls.model_fields.items()]
    
 

class TokenUsage(AbstractModel):
    """Tracks token usage for language model interactions"""
    model_config = {'protected_namespaces': ()}
    id: uuid.UUID| str  
    model_name: str
    tokens: typing.Optional[int] = Field(0,description="the number of tokens consumed in total")
    tokens_in: typing.Optional[int] = Field(0, description="the number of tokens consumed for input")
    tokens_out: typing.Optional[int] = Field(0, description="the number of tokens consumed for output")
    tokens_other: typing.Optional[int] = Field(0, description="the number of tokens consumed for functions and other metadata")
    session_id: typing.Optional[uuid.UUID| str  ] = Field(None, description="Session id for a conversation")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('tokens'):
            values['tokens'] = (values.get('tokens_in') or 0) + (values.get('tokens_out') or 0)
        return values
    
class IndexAudit(TokenUsage):
    model_config = {
        'index_notify': False        
    }
    """Track requests to build smart indexes such as graph links or text and image embeddings"""
    metrics: typing.Optional[dict] = Field(description="metrics for records indexed",default_factory=dict)
    status: str = Field(description="Status code such as OK|ERROR")
    message: typing.Optional[str] = Field(description="Any message such as an error")
    entity_full_name: str
    
    
class _OpenAIMessage(BaseModel):
    role: str
    content: typing.Optional[str] = Field('', description='text content')
    tool_calls: typing.Optional[typing.List[dict]|dict]
    tool_call_id: typing.Optional[str] = None
    
    @classmethod
    def from_message(cls, values):
        """
        
        """
        return _OpenAIMessage(**values, tool_call_id=(values.get('tool_eval_data') or {}).get('id'))
    
    @model_validator(mode='before')
    @classmethod
    def _val(cls,values):
        if tool_calls:= values.get('tool_calls'):
            if isinstance(tool_calls,dict):
                values['tool_calls'] = [tool_calls]
        return values

class AIResponse(TokenUsage):
    """Each atom in an exchange between users, agents, assistants and so on. 
    We generate questions with sessions and then that triggers an exchange. 
    Normally the Dialogue is round trip transaction.
    """
    model_config = {'protected_namespaces': ()}
    id: uuid.UUID| str  
    role: str = Field(description="The role of the user/agent in the conversation")
    content: str = DefaultEmbeddingField(description="The content for this part of the conversation") #TODO we may not want to automatically generate embeddings for this table
    status: typing.Optional[str] = Field(description="The status of the session such as REQUEST|RESPONSE|ERROR|TOOL_CALL|STREAM_RESPONSE")
    tool_calls: typing.Optional[typing.List[dict]|dict] = Field(default=None, description="Tool calls are requests from language models to call tools")
    tool_eval_data: typing.Optional[dict] = Field(default=None, description="The payload may store the eval from the tool especially if it is small data")
    verbatim: typing.Optional[dict|typing.List[dict]] = Field(default=None, description="the verbatim message from the language model - we dont serialize this", exclude=True)     
    function_stack: typing.Optional[typing.List[str]] = Field(None, description='At each stage certain functions are available to the model - useful to see what it has and what it chooses and to reload stack later')
 
    def to_open_ai_message(self):
        """the message structure for the scheme
        the one thing we need to do is make sure that tool calls reference the id(s)
        """
        return _OpenAIMessage.from_message(self.model_dump())
    
    @classmethod
    def from_open_ai_response(cls, response, sid:str)  :
        """"""
        choice = response['choices'][0]
        usage = response['usage']
        message = choice['message']
        tool_calls= message.get('tool_calls')
        return cls(
            id=str(uuid.uuid1()),
            role=message.get('role'),
            tool_calls=tool_calls,
            content=message.get('content') or '',
            tokens_in=usage['prompt_tokens'],
            tokens_out=usage['completion_tokens'],
            model_name=response['model'],
            status='TOOL_CALL' if not tool_calls else 'RESPONSE',
            session_id=sid
        )
class Session(AbstractModel):
    """Tracks groups if session dialogue - sessions are requests that produce actions or activity. 
    Typical example is interacting with an AI but also it could be simply a request to bookmark or ingestion some content.
    From the user's perspective a session is a single interaction but internally the system performs multiple hops in general - hence "session"
    Sessions can be interactions with the system or agent but they can be inferred e.g. from user activity
    Sessions can also be abstract for example a daily diary can be an abstract session with the intent of capturing daily activity
    Sessions can be hierarchical and moments can be generated from sessions.
    
    Sessions model user intent.
    """
    id: uuid.UUID| str  
    """i should maybe deprecate the name here as this is a dense audit but then maybe we should just archive?"""
    name: typing.Optional[str] = Field(None, description="The name is a pseudo name to make sessions node-compatible")
    query: typing.Optional[str] = DefaultEmbeddingField(None,description='the question or context that triggered the session - query can be an intent and it can be inferred')
    user_rating: typing.Optional[float] = Field(None, description="We can in future rate sessions to learn what works")
    agent: str = Field('percolate', description="Percolate always expects an agent but we support passing a system prompt which we treat as anonymous agent")
    parent_session_id:typing.Optional[uuid.UUID| str] = Field(None, description="A session is a thread from a question+prompt to completion. We span child sessions")
    
    thread_id: typing.Optional[str] = Field(None,description='An id for a thread which can contain a number of sessions - thread matches are case insensitive ie.. MyID==myid - typically we prefer ids to be uuids but depends on the system')
    channel_id: typing.Optional[str] = Field(None, description="The platform/channel ID through which the user is interacting (e.g., specific Slack channel ID)")
    channel_type: typing.Optional[str] = Field('percolate', description="The platform type (e.g., 'slack', 'percolate', 'email') - the device info + the channel type tells if its mobile etc")
    session_type: typing.Optional[str] = Field('conversation', description="The type of session (e.g., 'conversation', 'resource_management', 'task_creation', 'daily_digest')")

    metadata: typing.Optional[dict] = Field(default_factory=dict, description="Arbitrary metadata - typically this is device info and other context")
    session_completed_at: typing.Optional[datetime.datetime] = Field(default=None,description="An audit timestamp to write back the completion of the session - its optional as it should manage the last timestamp on the AI Responses for the session")
    graph_paths: typing.Optional[typing.List[str]] = Field(None, description="Track all paths extracted by an agent as used to build the KG over user sessions")
    userid:  typing.Optional[uuid.UUID| str ] = Field(None,description="The user id to use") #implicit
        
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values): 
        if isinstance(values,dict) and not values.get('id'):
            """the name is mapped for a hash of the query and can create collisions for now since we only want to index intent"""
            # TODO - consider if this is ok ^
            values['name'] = make_uuid({'query': values.get('query', '')})
        return values
    
    @classmethod
    def get_daily_diary_thread_id(cls, userid:str,  session_type:str=None):
        """this is a simple hash for filtering and respect the contract for threads"""
        daily_diary_id = make_uuid({"user_id": userid, "session_type": session_type})
        return daily_diary_id
    
    @classmethod
    def daily_diary_entry(cls, userid:str, query:str, metadata:dict=None, session_type:str='daily_digest'):
        """
        generates a task for the daily diary
        the query should be the intent e.g. a question from a user or an action such as upload file
        """
        
        return Session(id = str(uuid.uuid1()), 
                       query=query,
                       name =f"Diary entry {get_iso_timestamp()}",
                       session_type=session_type, 
                       userid=userid,
                       thread_id=cls.get_daily_diary_thread_id(userid=userid,session_type=session_type), 
                       metadata=metadata)
    
    @classmethod
    def task_thread_entry(cls, thread_id :str,  userid:str, query:str, metadata:dict=None,session_type:str='daily_digest'):
        """
        generates a task for the daily diary
        the query should be the intent e.g. a question from a user or an action such as upload file
        """
        
        return Session(id = str(uuid.uuid1()), 
                       query=query,
                       name =f"Diary entry {get_iso_timestamp()}",
                       session_type=session_type, 
                       userid=userid,
                       thread_id=thread_id, 
                       metadata=metadata)
    
    
    
    
    @classmethod
    def from_question_and_context(cls, id:str, question:str, context:typing.Any, agent:str=None, **kwargs):
        """
        from the session we can track the question and the agent that serviced. we may also want to store other details 
        """
        from percolate.services.llm import CallingContext
        context: CallingContext = context
        
        #TODO: we should capture some of context e.g. threads and users 
        return cls(query=question,
            id=id or str(uuid.uuid1()), 
            agent=agent,
            #TODO user id and channel contexts etc.
            **kwargs)
         
        
class SessionEvaluation(AbstractModel):
    """Tracks groups if session dialogue"""
    id: uuid.UUID| str  
    rating: float = Field(None,description="A rating from 0 to 1 - binary thumb-up/thumbs-down are 0 or 1")
    comments: typing.Optional[str] = Field(None, description="Additional feedback comments from the user")
    session_id: uuid.UUID| str  
    
class ModelMatrix(AbstractModel):
    """keep useful json blobs for model info"""
    id: uuid.UUID| str  
    entity_name: str = Field(description="The name of the entity e.g. a model in the types or a user defined model")
    enums: typing.Optional[dict] = Field(None, description="The enums used in the model - usually a job will build these and they can be used in query prompts")
    example_sql_queries: typing.Optional[dict] = Field(None, description="A mapping of interesting question to SQL queries - usually a job will build these and they can be used in query prompts")

class Category(AbstractEntityModel):
    """A category is just a general topic node that can be linked to other nodes and can have an evolving description or summary"""
    id: uuid.UUID| str  
    name: str = Field(description="The name of the category node")
    description: str = DefaultEmbeddingField(description="The content description for the category")
    
class Project(AbstractEntityModel):
    """A project is a broadly defined goal with related resources (uses the graph)"""
    id: uuid.UUID| str  
    name: str = Field(description="The name of the entity e.g. a model in the types or a user defined model")
    description: str = DefaultEmbeddingField(description="The content for this part of the conversation")
    target_date: typing.Optional[datetime.datetime] = Field(None, description="Optional target date")
    collaborator_ids: typing.List[uuid.UUID] = Field(default_factory=list, description="Users collaborating on this project")
    status: typing.Optional[str] = Field(default="active", description="Project status")
    priority: typing.Optional[int] = Field(default=1, description="Priority level (1-5), 1 being the highest priority")
    
class Task(Project):
    """Tasks are sub projects. A project can describe a larger objective and be broken down into tasks.
If if you need to do research or you are asked to search or create a research iteration/plan for world knowledge searches you MUST ask the _ResearchIteration agent_ to help perform the research on your behalf (you can request this agent with the help i.e. ask for an agent by name). 
You should be clear when relaying the users request. If the user asks to construct a plan, you should ask the agent to construct a plan. If the user asks to search, you should ask the research agent to execute  a web search.
If the user asks to look for existing plans, you should ask the research agent to search research plans.
    """
    id: typing.Optional[uuid.UUID| str] = Field(None,description= 'id generated for the name and project - these must be unique or they are overwritten')
    project_name: typing.Optional[str] = Field(None, description="The related project name if relevant")
    estimated_effort: typing.Optional[float] = Field(default=None, description="Estimated effort in hours")
    progress: typing.Optional[float] = Field(default=0.0, description="Completion progress (0-1)")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values): 
        if isinstance(values,dict) and not values.get('id'):
            values['id'] = make_uuid({'name': values['name'], 'project_name': values.get('project_name') or ''})
        return values
    
    @classmethod
    def get_model_functions(cls):
        """fetch task external functions"""
        return {'post_tasks_': 'Used to save tasks by posting the task object. Its good practice to first search for a task of a similar name before saving in case of duplicates' }
    
class User(AbstractEntityModel):
    """A model of a logged in user"""
    id: uuid.UUID| str  
    name: typing.Optional[str] = Field(None,description="A display name for a user")
    email: typing.Optional[str] = KeyField(None,description="email of the user if we know it - the key must be something unique like this")
    slack_id: typing.Optional[str] = Field(None,description="slack user U12345")
    linkedin: typing.Optional[str] = Field(None,description="linkedin profile for user discovery")
    twitter: typing.Optional[str] = Field(None,description="twitter profile for user discovery")

    description: typing.Optional[str] = DefaultEmbeddingField(default='', description="A learned description of the user")
    #current_chat_thread: typing.Optional[typing.List[str]] = Field(default_factory=list, description="A thread is a single conversation on any channel - the session contains a list of messages with thread id binding them")
    recent_threads: typing.Optional[typing.List[dict]|dict] = Field(default_factory=list, description="A thread is a single conversation on any channel - the session contains a list of messages with thread id binding them")
    last_ai_response: typing.Optional[str] = Field(None, description="We store this context for managing conversations and state")
    interesting_entity_keys: typing.Optional[dict] = Field(default_factory=dict, description="For the agents convenience a short term mapping of interesting keys with optional descriptions based on user activity")
    token: typing.Optional[str] = Field(None, description="A token for user authentication to Percolate")
    roles: typing.Optional[typing.List[str]] = Field(default_factory=list, description="A list of roles the user is a member of")
    graph_paths: typing.Optional[typing.List[str]] = Field(None, description="Track all paths extracted by an agent as used to build the KG")
    metadata: typing.Optional[dict] = Field(default_factory=dict, description="Arbitrary user metadata")
    email_subscription_active: typing.Optional[bool] = Field(False, description="Users can opt in and out of emails")
    
    @staticmethod
    def id_from_email(email):
        return make_uuid(email)
    
    def as_memory(self,**kwargs):
        """the user memory structure"""

        return {
            'Info': "You can use the users context - observe the current chat thread which may be empty when deciding if the user is referring to something they discussed recently or a new context."
                    "When you do use this context do not explain that to the user as it would be jarring for them. Freely use this context if its relevant or necessary to understand the user context."
                    "The last AI Response from the previous interaction is added for extra context and can be used if the user asks a follow up question in reference to previous response only. But dont ask them for confirmation."
                    "If entity keys are provided you can use the get-entities lookup function to load and inspect them.",
                    "some tools may accept a user id and you can use the user id here for those tool calls e.g. to do user specific searches"
                    
            'recent_threads': self.recent_threads,
            'last_ai_response': self.last_ai_response,
            'interesting_entity_keys': self.interesting_entity_keys,
            'users_name': self.name,
            'about user' : self.description,
            'user_id': self.id
        }
        
class UserFact(AbstractModel):
    """
    important information or trivial about a user typically submitted by a user over time.
    we can search specifically for user submitted data using vector search or entity lookup for labeled data.
    """
    id: uuid.UUID| str  
    name: typing.Optional[str] = Field(None,description="A unique entity name e.g. user_id-label")
    label: typing.Optional[str] = Field(None,description="The friendly label for the piece of information")
    description: typing.Optional[str] = DefaultEmbeddingField(default='', description="A description from the user e.g. my nick name is, my favourite color is, i have traveled to, i like foods like...")
    invalidated: typing.Optional[bool] = Field(False, description="We can store old information but invalidate it - if we retrieve such information we may discard it as relevant")
    graph_paths: typing.Optional[typing.List[str]] = Field(None, description="Track all paths extracted by an agent as used to build the KG")
    userid: str
    
    @classmethod
    def save_user_fact(cls, label:str, description:str, user_id: str, graph_paths:typing.List[str]=None):
        """save the user fact with a unique label for the information
        
        Args:
            description: details about the user fact
            unique_label: a user level unique label about the fact - should not collide with other user facts (best effort)
            user_id: the user id supplied in context- cif not known do not try to use this function
            graph_paths: graph paths are tags of the form A/B where A is more specific than B e.g. LLMs/AI
        """
        name = f"{user_id}.{label}"
        id = make_uuid(name)
        u = UserFact(id=id,name=name, label=label,description=description, graph_paths=graph_paths,userid=user_id)
        return p8.repository(u).update_records(u)
    
    @classmethod
    def search_user_facts(cls, query:str, user_id:str):
        """
        Search semantically filtered by a given user id
        
        Args:
            query: a detailed question to search for
            user_id: str the provided user id if known - if not known, do not try to use this function
        """
        
        """for now just search all but filter by user id assumed"""
        return p8.repository(UserFact).search(query,user_id=user_id)
    
    @classmethod
    def get_user_entity(cls, label:str,user_id:str):
        """looks up a user entity using the label qualified with the user id"""
        name = f"{user_id}.{label}"
        return p8.repository(UserFact).get_entities(name)
        
        
class Resources(AbstractModel):
    """Generic parsed content from files, sites etc. added to the system.
    If someone asks about resources, content, files, general information etc. it may be worth doing a search on the Resources.
    If a user expresses interests or preferences or trivia about themselves - you can save it in the background but response naturally in conversation about it.
    If the user asks information about themselves you can also try to search for user facts if you dont have the answer.
    """
    id: typing.Optional[uuid.UUID| str] = Field("The id is generated as a hash of the required uri and ordinal")  
    name: typing.Optional[str] = Field(None, description="A short content name - non unique - for example a friendly label for a chunked pdf document or web page title")
    category: typing.Optional[str] = Field(None, description="A content category")
    content: str = DefaultEmbeddingField(description="The chunk of content from the source")
    summary: typing.Optional[str] = Field(None,description="An optional summary")
    ordinal: int = Field(0, description="For chunked content we can keep an ordinal")
    uri: str = Field("An external source or content ref e.g. a PDF file on blob storage or public URI")
    metadata: typing.Optional[dict] = {} #for emails it could be sender and receiver info and date
    graph_paths: typing.Optional[typing.List[str]] = Field(None, description="Track all paths extracted by an agent as used to build the KG")
    resource_timestamp: typing.Optional[datetime.datetime] = Field(None,description= 'the database will add a default date but sometimes we want to control when the resource is relevant for - for example a daily log')
    userid:  typing.Optional[uuid.UUID| str] = Field(None, description="The user id is a system field but if we need to control the user context this is how we do it")  
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            """you need to be careful with the id i.e. ordinals and uris should make the resource unique
            filenames for example should not be used but replaced with e.g. an s3 global uri
            """
            values['id'] = make_uuid({'uri': values['uri'], 'ordinal': values.get('ordinal') or 0})
        return values
    
    @classmethod
    def search_facts_by_users(cls, query:str, user_id:str):
        """
        Search semantically filtered by a given user id
        
        Args:
            query: a detailed question to search for
            user_id: str the provided user id if known - if not known, do not try to use this function
        """
        
        return UserFact.search_user_facts(query, user_id)
    
    @classmethod
    def save_user_fact(cls, description:str, unique_label:str, user_id:str, graph_paths:typing.List[str]):
        """save the user fact with a unique label for the information
        
        Args:
            description: details about the user fact
            unique_label: a user level unique label about the fact - should not collide with other user facts (best effort)
            user_id: the user id supplied in context- if not known do not try to use this function
            graph_paths: graph paths are tags of the form A/B where A is more specific than B e.g. LLMs/AI
        """
        
        return UserFact.save_user_fact(unique_label, description=description,user_id=user_id, graph_paths=graph_paths)

    @classmethod
    def chunked_resource_from_text(cls,
        text: str,
        uri:str,
        chunk_size: int = 1000,
        category: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        userid: typing.Optional[str] = None
    ) -> typing.List["Resources"]:
        """load text into simple chunked resource"""
        name = name or uri
        chunks = [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]

        resources = []
        for i, chunk in enumerate(chunks):
            """TODO: NB!!!!!! HERE WE ARE USING URIS WITHOUT PARAMETERS AS GLOBALLY UNIQUE"""
            id_input = f"{uri.split('?')[0]}-{i}"
            id_hash = make_uuid(id_input)

            resource = cls(
                id=id_hash,
                name=f'{name} ({i})' if i > 0 else name,
                category=category,
                content=chunk,
                ordinal=i,
                uri=uri,
                userid=userid
            )
            resources.append(resource)

        return resources
        
        
    @classmethod
    def chunked_resource(
        cls,
        uri: str,
        chunk_size: int = 1000,
        category: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        userid: typing.Optional[str] = None,
        try_markdown:bool = False
    ) -> typing.List["Resources"]:
        """read file contents from web or file and chunks them by some chunk size - sensible defaults are used
        
        the uri is used as the id but we excluded parameters from the uri
        """
        
        from percolate.utils.parsing import get_content_provider_for_uri
        from pathlib import Path
        
        def sanitize_text(text: str) -> str:
            """this is a dump serializer for now as we work out the data model"""
            import html2text    
            t =  text.replace("\x00", "")  # or use a placeholder like "�"
            if try_markdown:
                try:
                     t = html2text.html2text(t)
                except Exception as ex:
                    #print("HTMLPARSE", ex)
                    pass
            return t
           
        """TODO add an S3 provider that understands signing""" 
        provider = get_content_provider_for_uri(uri)
        text = sanitize_text(provider.extract_text(uri))
        
        return cls.chunked_resource_from_text(text, 
                                              uri, 
                                              chunk_size=chunk_size,
                                              name=name,
                                              userid=userid, 
                                              category=category)

class TaskResources(AbstractModel):
    """(Deprecate)A link between tasks and resources since resources can be shared between tasks"""
    id: typing.Optional[uuid.UUID| str] = Field(None, description="unique id for rel" )  
    resource_id: uuid.UUID| str = Field("The resource id" )  
    session_id:  uuid.UUID| str = Field("The session id is typically a task or research iteration but can be any session id to group resources" )  
    user_metadata: typing.Optional[dict] = Field(default_factory=dict, description="User-specific metadata for this resource")
    relevance_score: typing.Optional[float] = Field(default=None, description="How relevant this resource is to the session")
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            """
            """
            values['resource_id'] = str(values['resource_id'])
            values['session_id'] = str(values['session_id'])
            
            values['id'] = make_uuid({'resource_id': values['resource_id'], 'session_id': values.get('session_id') or 0})
            
        return values
    
class SessionResources(AbstractModel):
    """A link between sessions and resources since resources can be shared between sessions"""
    id: typing.Optional[uuid.UUID| str] = Field(None, description="unique id for rel" )  
    resource_id: uuid.UUID| str = Field("The resource id" )  
    session_id:  uuid.UUID| str = Field("The session id is any user intent from a chat/request/trigger" )  
    count: typing.Optional[int] = Field(1, description="In a model where we only store head we can track an estimate for chunk count")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            """
            """
            values['resource_id'] = str(values['resource_id'])
            values['session_id'] = str(values['session_id'])
            values['id'] = make_uuid({'resource_id': values['resource_id'], 'session_id': values.get('session_id') or 0})
        return values
class _QuestionSet(BaseModel):
    query:str = Field(description="The question/query to search")
    concept_keys: typing.Optional[typing.List[str]] = Field(default_factory=list, description="A list of codes related to the concept diagram matching questions to the research structure")
    
class ResearchIteration(AbstractModel):
    """
    Your job is to use functions to execute research tasks. You do not search for answers directly, you post a research tasks to generate data.
    A research iteration is a plan to deal with a task.     
    If you are asked for a plan you should first use your json structure to create a plan and give it to the user.

    1. if you are asked to information on general topics, you should execute the post_tasks_research_execute
    2. if you are asked about other research iterations only then can you use the _search_ method which is designed to search research plans as opposed to actually search for general information.
    
    You can generate conceptual diagrams using mermaid diagrams to provide an overview of the research plan.
    When you generate a conceptual plan you should link question sets to plans for example each question should have labels that link to part of the conceptual diagram using the mermaid diagram format to describe your plan.
    """
    id: typing.Optional[uuid.UUID| str] = Field(None, description="unique id for rel" )  
    iteration: int
    content: typing.Optional[str] = DefaultEmbeddingField(None,description="An optional summary of the results discovered")
    conceptual_diagram: typing.Optional[str] = Field(None, description="The mermaid diagram for the plan - typically generated in advanced of doing a search")
    question_set: typing.List[_QuestionSet] = Field(description="a set of questions and their ids from the conceptual diagram")
    task_id: typing.Optional[uuid.UUID| str] = Field(default=None,description="Research are linked to tasks which are at minimum a question" )  
    
    @classmethod
    def get_model_functions(cls):
        """override model functions to provide my available behaviours/tools"""
        return {
            'post_tasks_research_execute': "post the ResearchIteration object to the endpoint to execute a research plan."
        }
        
    """we add some functions for illustration and testing but the database should be used for core functions"""
    # @classmethod
    # def perform_web_search(cls,query:str,limit:int=3)->typing.List[dict]:
    #     """performs a tavily web search - supply a query and receive a set of summaries and urls
        
    #     Args:
    #         query: the long form web query to search using the search api
    #         limit: how many search results to fetch
    #     """
        
    #     from percolate.utils.env import TAVILY_API_KEY
    #     import requests
        
    #     if not TAVILY_API_KEY:
    #         raise Exception(f"The TAVILY_API_KEY key is not set so we have no way to search")
        
    #     headers = { "Authorization": f"Bearer {TAVILY_API_KEY}"  }
        
    #     r  = requests.post("https://api.tavily.com/search", headers=headers, json={'query':query,'limit':limit})
    #     if not r.status_code in [200,201]:
    #         raise Exception(f"Error calling function - {r.content}")
    #     return r.json()
    

    
    @classmethod
    def fetch_web_content(cls,url:str, summarization_context:str=None):
        """provide a uri and fetch the web content - some pages not be accessible and you should typically report the error and try another resource
        
        Args:
            url: the web page to fetch - content will be converted to markdown
            summarization_context: it is recommended to pass a summarization context to reduce propagated tokens - the page will be summarized once in context
        """
        import html2text
        import requests
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"
        }
        
        data = requests.get(url, headers=headers).content.decode()
        data = html2text.html2text(data)
        if summarization_context:
            data = p8.summarize(data,summarization_context)
        return data
        
class BlockDocument(AbstractModel):
    """Generic parsed content from files, sites etc. added to the system"""
    id: typing.Optional[uuid.UUID| str] = Field("The id is generated as a hash of the required uri and ordinal")  
    name: typing.Optional[str] = Field(None, description="A short content name - non unique - for example a friendly label for a chunked pdf document or web page title")
    category: typing.Optional[str] = Field(None, description="A content category")
    summary: str = DefaultEmbeddingField(description="The chunk of content from the source")
    #conceptual blocks show document structure; inter block links, citations, entities referenced
    content_blocks:typing.List[dict] = Field(description="The content blocks in json format")
    conceptual_diagram: typing.Optional[str] = Field(None, description="The mermaid diagram for the plan or renewed for the document")
    metadata: typing.Optional[dict] = {} #metadata for a document
    
class PercolateAgent(Resources):
    """The percolate agent is the guy that tells you about Percolate which is a multi-modal database for managing AI in the data tier.
    You can learn about the philosophy of Percolate or ask questions about the docs and codebase.
    You can lookup entities of different types or plan queries and searches.
    You can call any registered apis and functions and learn more about how they can be used.
    Call the search function to get data about Percolate
    Only use 'search' if you are asked questions specifically about Percolate otherwise call for help!
    If you do a search and you do not find any data related to the user question you should ALWAYS ask for help to dig deeper.
    """
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        if not values.get('id'):
            values['id'] = make_uuid({'uri': values['uri'], 'ordinal': values['ordinal']})
        return values


class Settings(AbstractModel):
    """settings are key value pairs for percolate admin"""
    id: typing.Optional[uuid.UUID| str] = Field("The id is generated as a hash of the required key and ordinal. The id must be in a UUID format or can be left blank")  
    key: str = Field(default="The key for the value to store - id is generated form this")
    value: str = Field(description="Value of the setting")
    
    @model_validator(mode='before')
    @classmethod
    def _f(cls, values):
        values['id'] = make_uuid({"key": values['key']})
        return values
    
    @classmethod
    def inserts_from_map(cls, values:dict):
        """given some key value settings, generate an insert batch statement - might be a good idea to wrap these in try blocks """
        if not values:
            return ""

        statements = []
        for key, val in values.items():
            # Convert value to JSON-safe string, then escape for SQL
            escaped_key = json.dumps(key)
            sql_key = f"'{key}'"
            escaped_val = f"'{val}'"

            """we use a database function that hashes the key for the id and then we add the key value pair"""
            stmt = f"""
INSERT INTO p8."Settings" (id, key, value)
SELECT p8.json_to_uuid('{{"key": {escaped_key}}}'::JSONB), {sql_key}, {escaped_val}
ON CONFLICT (id)
DO UPDATE SET value = EXCLUDED.value;
""".strip()
            statements.append(stmt)

        return "\n".join(statements)
  
# Scheduled tasks model
class Schedule(AbstractModel):
    """Defines a scheduled task.
    Tasks are scheduled by the system or by users and can be a system prompt for an agent or a function call
    """
    id: typing.Optional[uuid.UUID| str] = Field(None, description="Unique schedule id")
    userid: typing.Optional[uuid.UUID| str] = Field(None, description="User id associated with schedule")
    name: str = Field(..., description="Task to execute")
    spec: dict = Field(..., description="the task spec - The task spec can be any json but we have an internal protocol. LLM instructions are valid should use a system_prompt attribute")
    schedule: str = Field(..., description="Cron schedule string, e.g. '0 0 * * *'")
    disabled_at: typing.Optional[datetime.datetime] = Field(None, description="Time when schedule was disabled")

    @model_validator(mode='before')
    @classmethod
    def _set_defaults(cls, values):
        if isinstance(values, dict) and not values.get('id'):
            values['id'] = str(uuid.uuid1())
        return values
 
 ###
 ## TODO: the functions on this object will move to SQL function and API level 
 ###
class Engram(AbstractModel):
    """Your role is to observe user text and then create a user memory.
    A user memory is a link between a `User` node named <user> and a <relationship> and a named concept.
    You should choose from canonical relationship names if you can;
    
    - only use knows relationships for person knows person
    - feel free to skip relationships that are not defined in the list below
    
    Dont assume strong relationships e.g. if someone is 'doing' something that does not mean they are 'skilled' in it.
    You should look for strong evidence to support each relationship or not mention it at all.........................
        
    **Relationship: **
    - `knows`
    - `likes`
    - `likes-not` 
    - `has-allergy` 
    - `has-goal`
    - `has-skill`
    - `working-on`
    - `is-learning`
    - `recommends`
    - `has-value`
    - `fears`
    - `takes-medication-for` 
    - `plans-to-attend` 
    - `advocates-for`
    - `opposes`
    - `teaches`
    - `aspires-to`
    - `reports-to`
    - 'observed'
    """

    @classmethod
    def _add_memory_from_user_sessions(cls, since_days_ago:int=1,limit:int=200):
        """
        A helper method to read the data from the sessions and generate memories
        
        Args:
            since_days_ago: how far back in days to look for updated session records
            limit: primarily not to overload models, limit the records. 
            Users are unlikely to have more than 100 sessions but file uploads count as sessions
            TODO: note this is currently for all users so we need to think about scale
        """
        
        from percolate.utils import get_days_ago_iso_timestamp 
        dt = get_days_ago_iso_timestamp(n=since_days_ago)
        
        """NOTE: we need to harden these types"""
        #in this case this filters things that are not based on user comments
        session_type ='conversation'
        
        data = p8.repository(Session).execute(f""" SELECT a.query, a.created_at, a.userid, a.updated_at, email, u.name
                                        FROM p8."Session" a
                                         join p8."User" u on a.userid = u.id 
                                        where a.updated_at > %s 
                                        and session_type = %s
                                        order by a.updated_at desc 
                                        limit {limit}""", data=(dt,session_type))
        
        logger.debug(f"Fetched {len(data)} records to convert to memories")
        
        if not data:
            return
        
        return p8.Agent(Engram).run(f"""Please add memories for the users by their name
                                    - the user_name comes from the email in the data 
                                    - check the function signature for parameter structure
                                    ```json
                                    {json.dumps(data,default=str)}
                                    ```
                                    """)
            
    @classmethod
    def add_memory(cls, relationships: typing.List[dict]):
        """
        Add a user memory by linking a user to a concept by a relationships type
        You should determine the user name from the content
        Te relationships are a list of dicts with `user_name`, `relationship` and `target_name`
        For target names try to use simple keywords and phrases without special characters or punctuation
        
        Args:
            relationships: a list of relationships object|dicts with `user_name`, `relationship` and `target_name`
        """
        
        from percolate.services import PostgresService
        
        for r in relationships:
            r['source_label'] = 'User'
            r['source_name'] = r['user_name']
            r['rel_type'] = r['relationship'].replace('-','_')
            
        r = PostgresService().graph.add_relationships(relationships)
                
        return r
    
    @classmethod
    def describe_user(cls, user_name:str):
        """supply a user name and get their memory graph to construct a narrative for that person 
        - feel free to infer some things the person might do or like but be clear if its assumed
        Make the description conversational rather than bullet points.
        Args:
            user_name: the unique user name as given
        """
        from percolate.services import PostgresService
         
        return PostgresService().graph.get_user_concept_links(user_name,as_model=True)
        

class AuditStatus(Enum):
    """Status enum for audit"""
    Fail="Fail"
    Success="Success"
class Audit(AbstractModel):
    """Generic event audit for Percolate"""
    id: typing.Optional[uuid.UUID| str] = Field(None, description="Unique id")
    userid: typing.Optional[uuid.UUID| str] = Field(None, description="Optional user context")
    status: str = Field(..., description="audit status e.g. Fail|Success")
    caller:str = Field(..., description="The calling object e.g. a class or service")
    status_payload: typing.Optional[dict] = Field(default_factory={}, description="Details about the event")
    error_trace : typing.Optional[str] = Field(None, description="If there is an error dump the stack trace")


# """The daily digest agent will evolve with the help of some API utils
# - we want to summarize changes in the user including their readings etc
# - we want to summarize new resources they have uploaded 
# - we want to specifically look at their evolving memory graph
# """
class DigestAgent(AbstractModel):
    """You are the digestion agent for the user - generate daily or weekly digests to email to the user.
    - Users upload resources like images, document, audio etc and we process them every day
    - Users also interact via that sessions
    - As users interact we build a knowledge graph for their memories. 
    Every time period you should look at what is changing and give a user summary. Create sections. 
    Only add the section when there is something relevant and do not comment on the fact that there is nothing to add if there is not.
    1. Provide a brief summary of the user; their goals, fears, thoughts and dreams as though you are taking to them like an oracle or good friend
    2. Do not be to nice to them. Be polite but challenge them to be better and to grow
    3. Call out any interesting information related to their goals
    4. Call out any appointments or reminders that came up that day that they may want to recall
    5. When looking at graph edges, observe any new edges added based on timestamps
    6. Call out an logical or factual inaccuracies or any conflicts or contradictions that they may want to
    7. Suggest any tasks they may want to keep an eye to get closer to their goals and any new tasks they may want to initiate (you may need access to their full task list)
    8. If you have access to their calendar (disabled for now) list any appointments coming up in the following day or two
    """
    @classmethod
    def get_daily_digest(cls,name:str, user_name:str,**kwargs):
        """
        The daily digest will be a combination of data feeds which could be pushed into the data tier
        It should be resource uploaded in the last N hours and the users profile
        """
        from percolate.utils import get_days_ago_iso_timestamp
        
        #fetch by the user name which is the email address
        user = p8.repository(User).execute(""" SELECT * FROM p8."User" where email = %s """, data=(user_name,))
        if not user:
            raise Exception(f"Could not find user by name `{user_name}`")
        if isinstance(user,list):
            user = user[0]
        user_id = user['id']
        
        #TODO we will unify user and p8 user
        q= f""" 'MATCH p = (u:User {{name: ''{user_name}''}})-[r*1..2]->(c:Concept)
                   RETURN u.name,  c.name as concept, relationships(p) as relationships  ',
                    'name agtype,  concept agtype, relationships agtype'"""
                                                    
        user_graph = p8.repository(User).execute(f"""   select * from cypher_query({q})  """)
        #TODO compact the graph
        
        """if would be good to render the graph as an image and make it available
           the caller can select the user_graph and do that if they wish or the agent can just build a summary from it
        """
        
        recent_resources_summarized = p8.repository(Resources).execute(""" SELECT * FROM p8."Resources" where userid=%s and updated_at >= %s  """,
                                                                            data=(user_id, get_days_ago_iso_timestamp(n=1)))
        
        """do a check on the context length for now and summarize if there is too much"""
        
        return {
            'user_last_session': user['updated_at'],
            'digest_scope': name,
            'user': user,
            'user_graph': user_graph,
            'recent_resources_upload': recent_resources_summarized            
        }


"""
Utility functions for the proxy module.

This module contains utility classes and functions for:
1. Background audit processing
2. Response parsing and formatting
3. Stream handling helpers
"""

import json
import threading
import typing
import uuid
import time
from pydantic import BaseModel
from percolate.utils import logger
from percolate.models import AIResponse, Session
import percolate as p8

class BackgroundAudit:
    """
    Background audit processor for AIResponse records.
    
    This sends AIResponse audit records on a background thread to avoid blocking
    the user stream. Used by higher level classes that construct their own AIResponse
    turns consisting of both tool calls from the LLM and local evaluation data.
    
    This auditor is designed to make streaming efficient while supporting full 
    auditing in the memory proxy.
    """
    
    def __init__(self):
        """Initialize the background auditor."""
        self._queue = []
        self._lock = threading.Lock()
        self._thread = None
        self._stop_event = threading.Event()
    
    def add_response(self, response: AIResponse) -> None:
        """
        Add an AIResponse to the audit queue.
        
        Args:
            response: The AIResponse to audit
        """
        with self._lock:
            self._queue.append(("ai_response", response))
            if not self._thread or not self._thread.is_alive():
                self._start_worker()
    
    def add_session(self, session_data: dict) -> None:
        """
        Add a session record to the audit queue.
        
        Args:
            session_data: The session data to audit
        """
        with self._lock:
            self._queue.append(("session", session_data))
            if not self._thread or not self._thread.is_alive():
                self._start_worker()
    
    def _start_worker(self) -> None:
        """Start the background worker thread."""
        self._stop_event.clear()
        self._thread = threading.Thread(target=self._process_queue, daemon=True)
        self._thread.start()
    
    def _process_queue(self) -> None:
        """Process the queue of audit records."""
        
        while not self._stop_event.is_set():
            with self._lock:
                if not self._queue:
                    break
                records = self._queue.copy()
                self._queue.clear()
            
            try:
                ai_responses = []
                session_records = []
                
                # Separate records by type
                for record_type, record in records:
                    if record_type == "ai_response":
                        ai_responses.append(record)
                    elif record_type == "session":
                        session_records.append(record)
                
                # Process AI responses
                if ai_responses:
                    p8.repository(AIResponse).update_records(ai_responses)
                    logger.debug(f"Audited {len(ai_responses)} AIResponse records")
                
                # Process sessions
                if session_records:
                    p8.repository(Session).update_records(session_records)
                    logger.debug(f"Audited {len(session_records)} Session records")
            except Exception as e:
                logger.error(f"Error processing audit queue: {e}")
            
            # Small sleep to avoid tight loop
            time.sleep(0.1)
    
    def stop(self) -> None:
        """Stop the background worker thread."""
        self._stop_event.set()
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout=2.0)
    
    
    def flush_ai_response_audit(
        self,
        content: str,
        tool_calls: typing.List[dict],
        tool_responses: typing.Dict[str, dict],
        usage: typing.Dict[str, int]
    ) -> None:
        """
        Flush AI response data to audit storage.
        
        This helper constructs an AI response and flushes it on a background thread.
        These are turns with request-response that can be used for auditing or memory.
        
        Args:
            content: The text content
            tool_calls: List of tool calls
            tool_responses: Dictionary of tool responses by tool call ID
            usage: Token usage information
        """

        ai_response = AIResponse(
            id=str(uuid.uuid4()),
            model_name="unknown",  # Required field
            role="assistant",
            content=content,
            status="TOOL_CALLS" if tool_calls else "RESPONSE",
            tool_calls=tool_calls,
            tool_eval_data=tool_responses,
            tokens_in=usage.get("prompt_tokens", 0),
            tokens_out=usage.get("completion_tokens", 0)
        )
        
        # Add to audit queue
        self.add_response(ai_response)
    
    @classmethod
    def audit_user_session(cls, 
                           session_id: str, 
                           user_id: str = None, 
                           channel_id: str = None,
                           query: str = None) -> None:
        """
        Audit a user session without creating an AIResponse.
        This method only creates or updates a Session record in the background.
        
        Args:
            session_id: The session ID
            user_id: Optional user ID
            channel_id: Optional channel ID
            query: Optional user query
        """
        # Create a session record
        session_data = {
            'id': session_id,
            'user_id': user_id,
            'channel_id': channel_id,
            'last_activity': time.time(),
            'query': query
        }
        
        # Add to the background auditor
        auditor = cls()
        auditor.add_session(session_data)


def parse_sse_line(line: str) -> typing.Optional[dict]:
    """
    Parse a server-sent event line into a JSON object.
    
    Args:
        line: The SSE line starting with 'data: '
        
    Returns:
        The parsed JSON object or None if parsing failed
    """
    if not line or not line.startswith("data: "):
        return None
    
    data = line[6:].strip()
    if data == "[DONE]":
        return {"type": "done"}
    
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        return None


def create_sse_line(data: typing.Union[dict, str]) -> str:
    """
    Create a server-sent event line from data.
    
    Args:
        data: The data to convert to an SSE line
        
    Returns:
        An SSE formatted line
    """
    if isinstance(data, dict):
        data = json.dumps(data)
    return f"data: {data}\n\n"


def format_tool_calls_for_openai(tool_calls: typing.List[dict]) -> typing.List[dict]:
    """
    Format tool calls to match the OpenAI API format.
    
    Args:
        tool_calls: List of tool calls in any format
        
    Returns:
        Tool calls in OpenAI format
    """
    openai_tool_calls = []
    
    for i, call in enumerate(tool_calls):
        # Extract function data
        if "function" in call:
            function = call["function"]
        else:
            # For Anthropic-style tool calls
            function = {
                "name": call.get("name", ""),
                "arguments": json.dumps(call.get("input", {}))
            }
        
        # Create OpenAI format tool call
        openai_tool_calls.append({
            "index": i,
            "id": call.get("id", f"call_{str(uuid.uuid4())[:16]}"),
            "type": "function",
            "function": {
                "name": function.get("name", ""),
                "arguments": function.get("arguments", "{}")
            }
        })
    
    return openai_tool_calls


"""
Example client functions for testing the admin content endpoints.
Requires the `requests` library: pip install requests
"""
import os
import requests
from percolate.utils import make_uuid
import json
import base64

# Base URL for Percolate API (override via environment)
BASE_URL = "http://127.0.0.1:5008"
# Retrieve test bearer token from environment
TOKEN = os.environ.get("P8_TEST_BEARER_TOKEN")
if not TOKEN:
    raise EnvironmentError("Environment variable P8_TEST_BEARER_TOKEN is not set.")

HEADERS = {"Authorization": f"Bearer {TOKEN}"}

device_info = base64.b64encode(json.dumps({'test':'device'}).encode('utf-8'))

def upload_uri_example(
    uri: str,
    description: str = "Sample bookmark for testing.",
    title: str = "Sample Bookmark",
    user_id: str = make_uuid('amartey@gmail.com'),
    device_info: str = device_info,
    is_public_resource: bool = True,
    expand_resource: bool = True,
    task_id: str = None,
):
    """
    Test the POST /admin/content/bookmark endpoint.

    Args:
        uri: URL to bookmark.
        description: Optional description or summary.
        title: Optional title of the resource.
        user_id: Optional user identifier (sent as query param).
        device_info: Optional base64-encoded JSON device metadata (query param).
        is_public_resource: Whether resource is public (query param).
        expand_resource: Whether to chunk the resource (query param).
        task_id: Optional session/thread ID (query param).
    """
    """
    Construct query parameters for optional args, send JSON body for URI data.
    """
    url = f"{BASE_URL}/admin/content/bookmark"
    headers = {**HEADERS, "Content-Type": "application/json"}
    # Body only contains URI and optional descriptive fields
    payload = {
        "uri": uri,
        "description": description,
        "title": title,
    }
    # Query params for session and metadata
    params = {}
    if user_id is not None:
        params['user_id'] = user_id
    if task_id is not None:
        params['task_id'] = task_id
    if device_info is not None:
        params['device_info'] = device_info
    params['is_public_resource'] = str(is_public_resource).lower()
    params['expand_resource'] = str(expand_resource).lower()
    # Send request
    response = requests.post(url, headers=headers, params=params, json=payload)
    # Debug info
    print("URL:", response.request.url)
    print("Request JSON:", payload)
    print("Query params:", params)
    print("Status code:", response.status_code)
    try:
        print("Response JSON:", response.json())
    except ValueError:
        print("Response text:", response.text)
    return response

def upload_file_example(
    file_path: str,
    task_id: str = None,
    add_resource: bool = True,
    user_id: str = make_uuid('amartey@gmail.com'),
    device_info: str = device_info,
):
    """
    Test the POST /admin/content/upload endpoint.

    Args:
        file_path: Path to local file to upload.
        task_id: Session/task identifier.
        add_resource: Whether to index the file as a resource.
        user_id: Optional user identifier.
        device_info: Optional base64 device info string.
    """
    url = f"{BASE_URL}/admin/content/upload"
    headers = HEADERS.copy()
    

    # 'requests' sets appropriate multipart/form-data headers
    with open(file_path, "rb") as f:
        files = {"file": (os.path.basename(file_path), f)}
        data = {
            "task_id": task_id,
            "add_resource": str(add_resource).lower(),
            "user_id": user_id or None,
            "device_info": device_info or None,
        }
        response = requests.post(url, headers=headers, files=files,  params=data)
    print("Uploaded file:", file_path)
    print("Request data:", data)
    print("Status code:", response.status_code)
    try:
        print("Response JSON:", response.json())
    except ValueError:
        print("Response text:", response.text)
    return response

if __name__ == "__main__":
    # Example URI bookmark
    upload_uri_example("https://example.com")
    # Example file upload: adjust sample.txt path if necessary
    sample_file = os.path.join(os.path.dirname(__file__), "sample.txt")
    if os.path.exists(sample_file):
        upload_file_example(sample_file)
    else:
        print(f"Sample file not found at {sample_file}, skipping file upload example.")


"""wrap all language model apis - use REST direct to avoid deps in the library
This is a first draft - will map this to lean more on the database model 
"""

import requests
import json
import os
import typing
from .CallingContext import CallingContext
from percolate.models import MessageStack
from percolate.services import PostgresService
from percolate.models.p8 import AIResponse
import uuid
from percolate.utils import logger
import traceback
from .MessageStackFormatter import MessageStackFormatter
from .utils import *


ANTHROPIC_MAX_TOKENS_IN = 8192

class OpenAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str,  model_name:str, streaming_callback:typing.Callable=False)->AIResponse:
        """
        parse the response into our canonical format - if streaming callback send to print
        
        example tool call response
        ```
        'tool_calls': [{'id': 'call_0KPgsQaaso8IXPUpG6ktM1DC',
        'type': 'function',
        'function': {'name': 'get_weather',
            'arguments': '{"city":"Dublin","date":"2023-10-07"}'}}],
        ```
        """
        def adapt(t):
                """we want something we can call and also something to construct the message that is needed for the tool call"""
                f = t['function']
                return   {'name': f['name'], 'arguments':f['arguments'], 'id': t['id']} 
            
        try:
            if streaming_callback:
                logger.debug(f"Streaming response")
                response = stream_openai_response(response,printer=streaming_callback)
            else:
                response = response.json()
            
           
            if response.get('error'):
                logger.warning(f"Error response {response['error'].get('message')}")
                return AIResponse(id = str(uuid.uuid1()),model_name=model_name, tokens_in=0,tokens_out=0, role='assistant',
                                session_id=sid, content=response['error'].get('message'), status = 'ERROR')
            choice = response['choices'][0]
            tool_calls = choice['message'].get('tool_calls') or []
            tool_calls = [adapt(t) for t in tool_calls]
            
            return AIResponse(id = str(uuid.uuid1()),
                    model_name=response['model'],
                    tokens_in=response['usage']['prompt_tokens'],
                    tokens_out=response['usage']['completion_tokens'],
                    session_id=sid,
                    verbatim=choice['message'],
                    role=choice['message']['role'],
                    content=choice['message'].get('content') or '',
                    status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                    tool_calls=tool_calls)
        except Exception as ex:
            logger.warning(f"unexpected structure in OpenAI scheme message {response=} - caused the error {ex}")
            raise 
                    
class AnthropicAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str,  model_name:str, streaming_callback:typing.Callable=False )->AIResponse:
        """parse the response into our canonical format - if streaming callback send to print"""
        if streaming_callback:
            response = stream_anthropic_response(response,printer=streaming_callback)
        else:
            response = response.json()
        choice = response.get('content') or []
        def adapt(t):
            """anthropic map to our interface"""
            return {'name': t['name'], 'arguments':t['input'], 'id': t['id'], 'scheme': 'anthropic'}
        verbatim = [t for t in choice if t['type'] == 'tool_use']
        tool_calls = [adapt(t) for t in verbatim]
        if verbatim:
            """when tools are used we need a verbatim message with tool call??"""
            verbatim = {
                'role': response['role'],
                'content': response['content']
            }
        
        content = "\n".join([t['text'] for t in choice if t['type'] == 'text']) 
        return AIResponse(id = str(uuid.uuid1()),
                model_name=response['model'],
                tokens_in=response['usage']['input_tokens'],
                tokens_out=response['usage']['output_tokens'],
                session_id=sid,
                role=response['role'],
                content=content or '',
                verbatim=verbatim ,
                status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                tool_calls=tool_calls)
        
class GoogleAIResponseScheme(AIResponse):
    @classmethod
    def parse(cls, response:requests.models.Response, sid: str, model_name:str, streaming_callback:typing.Callable=False)->AIResponse:
        """parse the response into our canonical format - if streaming callback send to print"""
        if streaming_callback:
            response = stream_google_response(response,printer=streaming_callback)
        else:
            response= response.json()
        message = response['candidates'][0]['content']
        choice = message['parts']
        content_elements = [p['text'] for p in choice if p.get('text')]
        def adapt(t):
            return {'function': {'name': t['name'], 'arguments':t['args'], 'id': t['name'],'scheme': 'google'}}
        tool_calls = [adapt(p['functionCall']) for p in choice if p.get('functionCall')]
        tool_calls = [t['function'] for t in tool_calls]
        return AIResponse(id = str(uuid.uuid1()),
                model_name=model_name, #does not seem to return it which is fair
                tokens_in=response['usageMetadata']['promptTokenCount'],
                tokens_out=response['usageMetadata']['candidatesTokenCount'],
                session_id=sid,
                role=message['role'],
                content=',\n'.join(content_elements),
                verbatim=message if tool_calls else None,
                status='RESPONSE' if not tool_calls else "TOOL_CALLS",
                tool_calls=tool_calls)

def try_get_open_ai_key():
    """
    for now we default to open ai keys for many types of models e.g. whisper or embeddings but we could generally use this to load other keys but we need to flesh out the abstraction
    """
    db = PostgresService()
    
    try:
        params = db.execute("""select token from p8."LanguageModelApi" where token_env_key = 'OPENAI_API_KEY' and token is not null limit 1""")
        if params:
            return params[0]['token']
    except:
        logger.warning(f"failed to get the open ai key - {traceback.format_exc()}")        
    

class LanguageModel:
    """the simplest language model wrapper we can make"""
    def __init__(self, model_name:str):
        """"""
        self.model_name = model_name
        self.db = PostgresService()
        #TODO we can use a cache for this in future
        self.params = self.db.execute('select * from p8."LanguageModelApi" where name = %s ', (model_name,))
        if not self.params:
            raise Exception(f"The model {model_name} does not exist in the Percolate settings")
        self.params = self.params[0]
        
        if self.params['token'] is None:
            """if the token is not stored in the database we use whatever token env key to try and load it from environment"""
            self.params['token'] = os.environ.get(self.params['token_env_key'])
            if not self.params['token']:
                raise Exception(f"There is no token or token key configured for model {self.model_name} - you should add an entry to Percolate for the model using the examples in p8.LanguageModelApi")
        """we use the env in favour of what is in the store"""
        env_token = os.environ.get(self.params['token_env_key'])
        self.params['token'] = env_token  if env_token is not None and len(env_token) else self.params.get('token')
        self._scheme = self.params.get('scheme','openai')
        
    def parse(self, response: requests.models.Response | typing.Any, context: CallingContext=None) -> AIResponse:
        """the llm response or HybridResponse from streaming is parsed into our canonical AIResponse.
        this is also done inside the database and here we replicate the interface before dumping and returning to the executor
        """
 
        streaming_callback = context.streaming_callback if context and context.streaming_callback else None

        try:
            if response.status_code not in [200,201]:
                logger.warning(f"There was an error requesting - {response.content}")
                pass #do something for errors
 
            sid = None if not context else context.session_id
            """check the HTTP response first"""
            if self.params.get('scheme') == 'google':
                return GoogleAIResponseScheme.parse(response, sid=sid, model_name=self.model_name,streaming_callback=streaming_callback)
            if self.params.get('scheme') == 'anthropic':
                return AnthropicAIResponseScheme.parse(response,sid=sid,model_name=self.model_name,streaming_callback=streaming_callback)
            return OpenAIResponseScheme.parse(response,sid=sid, model_name=self.model_name,streaming_callback=streaming_callback)
        except Exception as ex:
            logger.warning(f"failing to parse {response} {traceback.format_exc()}")
            raise
        
    def __call__(self, messages: MessageStack, functions: typing.List[dict], context: CallingContext=None, debug_response:bool=False ) -> AIResponse:
        """call the language model with the message stack and functions.
        We return the parsed response unless the caller asks for the raw
        We support a hybrid mode of return the content stream and holding the function call
        """
            
        try:
            response = self._call_raw(messages=messages, functions=functions,context=context)
            
            logger.debug(f"{response=}, {context=}")
            if debug_response:
                return response
            
       
            
            """for consistency with DB we should audit here and also format the message the same with tool calls etc."""
            response = self.parse(response,context=context)
            return response
        except:
            # data = {
            #     'messages': messages,
            #     'functions': functions,
            #     'context': context
            # }
            
            # import pickle
            # with open('/tmp/p8dump', "wb") as f:
            #     pickle.dump(data, f)
            
            # print('dumped', '/tmp/p8dump')
            raise
    
    def ask(self, question:str, functions: typing.List[dict]=None, system_prompt: str=None,context: CallingContext=None, **kwargs):
        """simple check frr question. our interface normally uses MessageStack and this is a more generic way
        
        If you want to stream create a context 
        Args:
            question: any question
            system_prompt: to test altering model output behaviour
            functions: optional list of functions in the OpenAPI like scheme
        """
        return self.__call__(MessageStack(question,system_prompt=system_prompt), functions=functions, context=context, **kwargs)
        
        
    def _call_raw(self, messages: MessageStack, functions: typing.List[dict], context: CallingContext=None):
         """the raw api call exists for testing - normally for consistency with the database we use a different interface"""
   
         return self.call_api_simple(messages.question, 
                                    functions=functions,
                                    system_prompt=messages.system_prompt, 
                                    data_content=messages.data,
                                    is_streaming=(context and context.is_streaming))

    @classmethod 
    def from_context(cls, context: CallingContext) -> "LanguageModel":
        return LanguageModel(model_name=context.model)
        
    def get_embedding(self, text: str) -> typing.List[float]:
        """Get embedding for a single text
        
        Args:
            text: Text to embed
            
        Returns:
            Embedding vector as list of floats
        """
        from percolate.utils.embedding import get_embedding
        
        # Use the embedding utility with our model settings
        scheme = self.params.get('scheme', 'openai')
        model = self.params.get('embedding_model', self.model_name)
        
        return get_embedding(
            text=text,
            model=model,
            api_key=self.params['token'],
            scheme=scheme
        )
        
    def get_embeddings(self, texts: typing.List[str]) -> typing.List[typing.List[float]]:
        """Get embeddings for multiple texts in batch
        
        This is much more efficient than calling get_embedding multiple times
        as it batches the requests to the embedding provider.
        
        Args:
            texts: List of texts to embed
            
        Returns:
            List of embedding vectors
        """
        from percolate.utils.embedding import get_embeddings
        
        # Use the embedding utility with our model settings
        scheme = self.params.get('scheme', 'openai')
        model = self.params.get('embedding_model', self.model_name)
        
        return get_embeddings(
            texts=texts,
            model=model,
            api_key=self.params['token'],
            scheme=scheme
        )
    
      
    def _elevate_functions_to_tools(self, functions: typing.List[dict]):
        """dialect of function wrapper for openai scheme tools"""
        return [{'type': 'function', 'function': f} for f in functions or []]
          
    def _adapt_tools_for_anthropic(self, functions: typing.List[dict]):
        """slightly different dialect of function wrapper - rename parameters to input_schema"""
        def _rewrite(d):
            return {
                'name' : d['name'],
                'input_schema': d['parameters'],
                'description': d['description']
            }
 
        return [_rewrite(d) for d in functions or []]
    
    def call_api_simple(self, 
                        question:str, 
                        functions: typing.List[dict]=None, 
                        system_prompt:str=None, 
                        data_content:typing.List[dict]=None,
                        is_streaming:bool = False,
                        temperature: float = 0.0,
                        **kwargs):
        """
        Simple REST wrapper to use with any language model
        """
        logger.debug(f"invoking model {self.model_name}, {is_streaming=}")
        """select this from the database or other lookup
        e.g. db.execute('select * from "LanguageModelApi" where name = %s ', ('gpt-4o-mini',))[0]
        """
        params = self.params
        data_content = data_content or []
        
        """we may need to adapt this e.g. for the open ai scheme"""
        tools = functions or None
      
        url = params["completions_uri"]
        """use the env first"""
         
        token = params['token']
        if not token or len(token)==0:
            raise Exception(f"There is no API KEY in the env or database for model {self.model_name} - check ENV {params.get('token_env_key')}")
        headers = {
            "Content-Type": "application/json",
        }
        if params['scheme'] == 'openai':
            headers["Authorization"] = f"Bearer {token}"
            tools = self._elevate_functions_to_tools(functions)
        if params['scheme'] == 'anthropic':
            headers["x-api-key"] = token
            headers["anthropic-version"] = self.params.get('anthropic-version', "2023-06-01")
            tools = self._adapt_tools_for_anthropic(functions)
        if params['scheme'] == 'google':
            url = f"{url}?key={token}" if not is_streaming else f"{url.replace('generateContent','streamGenerateContent')}?alt=sse&key={token}"
        data = {
            "model": params['model'],
            "messages": [
                *[{'role': 'system', 'content': s} for s in [system_prompt] if s],
                {"role": "user", "content": question},
                #add in any data content into the stack for arbitrary models
                *data_content
            ],
            "tools": tools,
            'temperature': temperature
        }
        if is_streaming:
            data['stream'] = True
            data["stream_options"] = {"include_usage": True}
            
        if params['scheme'] == 'anthropic':
            data = {
                "model": params['model'],
                'temperature': temperature,
                "messages": [
                    {"role": "user", "content": question}, 
                    #because they use blocks https://docs.anthropic.com/en/docs/build-with-claude/tool-use#example-of-empty-tool-result
                    *[MessageStackFormatter.adapt_tool_response_for_anthropic(d) for d in data_content if d]
                ]
            }
            if tools:
                data['tools'] = tools
            if system_prompt:
                data['system'] = system_prompt
            if is_streaming:
                data['stream'] = True
                
            data["max_tokens"] = kwargs.get('max_tokens',ANTHROPIC_MAX_TOKENS_IN)
             
        if params['scheme'] == 'google':
            data_content = [MessageStackFormatter.adapt_tool_response_for_google(d) for d in data_content if d]
            optional_tool_config ={}#{ "tool_config": {   "function_calling_config": {"mode": "ANY"}  }  } #this seems to confuse the googs
            data = {
                "contents": [
                    {"role": "user", "parts": [{'text': question}]},
                    *data_content
                ],
                "tools": [{'function_declarations': tools}] if tools else None,
            }
            """gemini is stupid if you add a tool to use but it already has the answer
            "role": "user",  "parts": [{   "functionResponse": { <--- if the message is like this disable tools
            """
            if not data_content or 'functionResponse' not in data_content[-1]['parts'][0]:
                data.update(optional_tool_config)
            if system_prompt:
                data["system_instruction"] =  { "parts": { "text": system_prompt } }
                    
        logger.trace(f"request {data=}, {is_streaming=}")
   
        response =  requests.post(url, headers=headers, data=json.dumps(data), stream=is_streaming)
        
        if response.status_code not in [200,201]:
            logger.warning(f"failed to submit: {response.status_code=}  {response.content}")
        if response.status_code == 429:
            import time
            """TODO: should be a flagged thing"""
            time.sleep(61)
            logger.warning(f"RATE LIMITED - SLEEPING FOR 1 MINUTE")
            response =  requests.post(url, headers=headers, data=json.dumps(data),stream=is_streaming)   
        return response
    
    
    
    def parse_ai_response(self, partial: dict, usage: dict, ctx: CallingContext) -> AIResponse:
        """
        Normalize a partial LLM response (tool call or content) and usage dict into an AIResponse.
        
        This method creates an AIResponse object, then optionally audits it using the background worker.
        
        Args:
            partial: Partial response data (content or tool calls)
            usage: Token usage dictionary
            ctx: Calling context with session info
            
        Returns:
            AIResponse: The created response object
        """
        # Normalize the incoming payload to a plain dict
        if hasattr(partial, 'model_dump'):
            data = partial.model_dump()
        elif hasattr(partial, 'dict'):
            try:
                data = partial.dict()
            except Exception:
                data = {}
        elif isinstance(partial, dict):
            data = partial
        else:
            data = {}

        # Extract core fields
        content = data.get('content', '')
        role = data.get('role', 'assistant')
        tool_calls = data.get('tool_calls')
        tool_eval_data = data.get('tool_eval_data') or (data if tool_calls else None)
        verbatim = data.get('verbatim')
        function_stack = data.get('function_stack')

        # Token usage comes solely from the provided usage dict
        if usage and isinstance(usage, dict):
            tokens_in = usage.get('prompt_tokens', usage.get('input_tokens', 0))
            tokens_out = usage.get('completion_tokens', usage.get('output_tokens', 0))
        else:
            tokens_in = tokens_out = 0

        # Determine status based on presence of a tool call
        status = 'TOOL_CALL_RESPONSE' if tool_calls else 'COMPLETED'

        # Create the AIResponse object
        ai_response = AIResponse(
            id=str(uuid.uuid1()),
            model_name=self.model_name,
            tokens_in=tokens_in,
            tokens_out=tokens_out,
            session_id=ctx.session_id,
            role=role,
            content=content,
            status=status,
            tool_calls=tool_calls,
            tool_eval_data=tool_eval_data,
            verbatim=verbatim,
            function_stack=function_stack,
        )
        
        # If we're configured to audit, send to background worker
        if getattr(ctx, 'audit', True):
            self._audit_response(ai_response)
            
        return ai_response
        
    def _audit_response(self, response: AIResponse) -> None:
        """
        Send an AIResponse to the background audit worker.
        
        Args:
            response: The AIResponse to audit
        """
        try:
            from percolate.services.llm.proxy.utils import BackgroundAudit
            
            # Get or create a background auditor
            if not hasattr(self, '_auditor'):
                self._auditor = BackgroundAudit()
                
            # Send to background worker
            self._auditor.add_response(response)
            
        except Exception as ex:
            logger.warning(f"Failed to audit AIResponse: {ex}")
            # Don't raise - auditing failures shouldn't block the response flow
    
    def stream_with_proxy(self, messages, functions, context):
        """
        Stream LLM requests using the proxy architecture.
        
        This method handles the lower-level details of creating the appropriate
        request object and passing it to the proxy stream generator.
        
        Args:
            messages: MessageStack or list of message dictionaries
            functions: Function definitions to include in the request
            context: CallingContext with API credentials and settings
        
        Returns:
            Generator yielding (line, chunk) tuples
        """
        from percolate.services.llm.proxy.stream_generators import request_stream_from_model
        from percolate.services.llm.proxy.models import OpenAIRequest, AnthropicRequest, GoogleRequest
        
        message_data = messages.data if hasattr(messages, 'data') else messages
        
        R = OpenAIRequest
        if self._scheme == 'anthropic':
            R = AnthropicRequest 
        elif self._scheme == 'google':
            R = GoogleRequest 
         
        request = R(
            model=self.model_name,
            messages=message_data,
            stream=True,
            tools=functions
        )
        
        # Use proxy stream generator - source_scheme determined by model params
        return request_stream_from_model(
            request=request,
            context=context,
            target_scheme='openai',  
            relay_tool_use_events=False,  
            relay_usage_events=False  
        )
        
    def get_stream_iterator(self, content_generator, context:CallingContext):
        """
        The stream iterator is a wrapper that helps with custom streaming logic, formatting and auditing
        """
        # Wrap the streaming generator and collected AIResponses for auditing
        return LLMStreamIterator(
            content_generator,
            scheme=self._scheme,
            context=context
        )
        


 
from fastapi import APIRouter, HTTPException, Query, Path, Response,BackgroundTasks
from percolate.models.p8 import Task,ResearchIteration
import percolate as p8
from percolate.api.routes.auth import get_current_token
import uuid
from fastapi import   Depends
import typing
from pydantic import BaseModel
from percolate.utils import logger

router = APIRouter()

@router.get("/")
async def get_tasks(user: dict = Depends(get_current_token))->typing.List[Task]:
    return Response('Dummy response')

class TaskSearch(BaseModel):
    query: str

@router.post("/search", response_model=typing.List[Task])
async def search_task(search: TaskSearch, user: dict = Depends(get_current_token)) -> typing.List[Task]:
    """semantic task search"""
    result = p8.repository(Task).search(search.query)
    
    """the semantic result is the one we want here"""
    if result and result[0].get('vector_result'):
        return result[0]['vector_result']
    
    """todo error handling"""
    
@router.post("/", response_model=Task)
async def create_task(task: Task, user: dict = Depends(get_current_token)) -> Task:
    """create a task"""
    results = p8.repository(Task).update_records(task)
    if results:
        return results[0]
    raise Exception("this should not happened but we will be adding error stuff")

@router.get("/{task_name}/comments")
async def get_task_comments_by_name(
    task_name: str = Path(..., description="The unique name of the task"),
    user: dict = Depends(get_current_token)
) -> typing.List[dict]:
    """Fetch the comments related to this task if you know its entity name"""
    return [{
        'user': 'dummy_user',
        'comment': 'dummy_comment'
    },{
        'user': 'dummy_user',
        'comment': 'dummy_comment'
    }]

@router.get("/{task_name}", response_model=Task)
async def get_task_by_name(
    task_name: str = Path(..., description="The unique name of the task"),
    user: dict = Depends(get_current_token)
) -> Task:
    """Retrieve a task by name"""
    return {}

@router.post("/research", response_model=ResearchIteration)
async def create_research_iteration(
    task: ResearchIteration,
    user: dict = Depends(get_current_token)
) -> ResearchIteration:
    """create a research plan - conceptual diagram and question set"""
    results = p8.repository(ResearchIteration).update_records(task)
    if results:
        return results[0]
    raise Exception("this should not happened but we will be adding error stuff")

def exec_research_tasks(task):
    repo = p8.repository(ResearchIteration)
    """this can be parallel in future but now our free tier is one request per second anyway"""
    for q in task.question_set:
        """this guy does a lot
        - first we do the web search and fetch the content optionally 
        - we insert the results into task/resources
        - we build the index - a smart index can also fetch content when the first step does not
        - the index is a Percolate special that adds vector and graph index for parsed content
        """
        repo.execute(f"SELECT * FROM p8.insert_web_search_results(%s,%s)", data=(q.query,task.task_id))
        
    return {}
            
@router.post("/research/execute", response_model=ResearchIteration)
async def execute_research_iteration(
    task: ResearchIteration,
    user: dict = Depends(get_current_token)
) -> ResearchIteration:
    """execute a research plan - perform each search in the question set - this can take time so should be done as background tasks"""
    
    logger.debug(task)
    
    result = exec_research_tasks(task)
    
    """handle errors and/or update the task response"""
    return task

@router.post("/research/queue", response_model=ResearchIteration)
async def queue_research_iteration(
    task: ResearchIteration, 
    background_tasks: BackgroundTasks,
    user: dict = Depends(get_current_token)
) -> ResearchIteration:
    """execute a research plan - perform each search in the question set - this can take time so should be done as background tasks"""
    
    result = background_tasks(exec_research_tasks, task)
    
    """handle errors and/or update the task response to say we have changed its status to queue"""
    
    return task
        
    



# @router.put("/{task_name}")
# async def update_task(task_name: str, task: Task):
#     pass

# @router.delete("/{task_id}")
# async def delete_task(draft_id: uuid.UUID):
#     pass



from pydantic import BaseModel
from percolate.models.AbstractModel import ensure_model_not_instance, AbstractModel
from percolate.models.utils import SqlModelHelper
from percolate.utils import logger, batch_collection
import typing
from percolate.models.p8 import ModelField, Agent, IndexAudit
import psycopg2
from percolate.utils.env import (
    POSTGRES_CONNECTION_STRING,
    POSTGRES_DB,
    POSTGRES_SERVER,
    DEFAULT_CONNECTION_TIMEOUT,
)
import psycopg2.extras
from psycopg2 import sql
from psycopg2.errors import DuplicateTable
from tenacity import retry, stop_after_attempt, wait_fixed
import traceback
import uuid
import json
from percolate.models.p8 import Function
from percolate.services.PercolateGraph import PercolateGraph


class PostgresService:
    """the postgres service wrapper for sinking and querying entities/models"""

    def __init__(
        self,
        model: BaseModel = None,
        connection_string=None,
        on_connect_error: str = None,
    ):
        try:
            self._connection_string = connection_string or POSTGRES_CONNECTION_STRING
            self.conn = None
            self._graph = PercolateGraph(self)
            self.helper = SqlModelHelper(AbstractModel)
            if model:
                """we do this because its easy for user to assume the instance is what we want instead of the type"""
                self.model = AbstractModel.Abstracted(ensure_model_not_instance(model))
                self.helper: SqlModelHelper = SqlModelHelper(model)
            else:
                self.model = None
            self.conn = psycopg2.connect(self._connection_string)

        except:
            if on_connect_error != "ignore":
                logger.warning(traceback.format_exc())
                logger.warning(
                    "Could not connect - you will need to check your env and call pg._connect again"
                )

    @property
    def graph(self):
        return self._graph

    def _create_db(self, name: str):
        """this util is to create a test database primarily"""

        if not self.conn:
            raise Exception(
                "The connection was not established - check the connection string and db service"
            )
        self.conn.autocommit = True
        try:
            cursor = self.conn.cursor()
            cursor.execute(f"""SELECT 1 FROM pg_database WHERE datname = '{name}';""")
            exists = cursor.fetchone()

            if not exists:
                logger.info(f"Creating database: {name}")
                cursor.execute(f"CREATE DATABASE {name};")
            else:
                logger.debug(f"Database {name} already exists.")

            cursor.close()
            self.conn.close()
        except Exception as ex:
            logger.debug(ex)
            raise
        finally:
            pass

    def __repr__(self):
        return f"PostgresService({self.model.get_model_full_name() if self.model else None}, {POSTGRES_SERVER=}, {POSTGRES_DB=})"

    def _reopen_connection(self):
        """util to retry opening closed connections in the service"""

        @retry(wait=wait_fixed(1), stop=stop_after_attempt(4), reraise=True)
        def open_connection_with_retry(conn_string):
            return psycopg2.connect(conn_string, connect_timeout=5)

        try:
            if self.conn is None:
                self.conn = open_connection_with_retry(self._connection_string)
            self.conn.poll()
        except psycopg2.InterfaceError as error:
            self.conn = None  # until we can open it, lets not trust it
            self.conn = open_connection_with_retry(self._connection_string)

    def _connect(self):
        self.conn = psycopg2.connect(self._connection_string)
        return self.conn

    @property
    def entity_exists(self):
        """convenience to see if the entity exists"""

        return self.check_entity_exists()

    def check_entity_exists(self):
        """sanity check for tets"""

        assert self.model, "trying to check exists when model is null is not allowed"
        Q = """SELECT EXISTS (    SELECT FROM information_schema.tables    WHERE table_schema = %s AND table_name = %s   )"""
        result = self.execute(
            Q, data=(self.model.get_model_namespace(), self.model.get_model_name())
        )
        if result:
            return result[0]["exists"]
        return False

    def repository(self, model: BaseModel, **kwargs) -> "PostgresService":
        """a connection in the context of the abstract model for crud support"""
        return PostgresService(
            model=model, connection_string=self._connection_string, **kwargs
        )

    def get_entities(self, keys: str | typing.List[str], userid: str = None):
        """
        use the get_entities database function to lookup entities, with optional user_id for access control

        Args:
            keys: one or more business keys (list of entity names) to fetch
            userid: optional user identifier to include private entities owned by this user
        """
        if keys:
            if not isinstance(keys, list):
                keys = [keys]
        # Call SQL function with optional userid; when userid is None, only public entities are returned
        data = (
            self.execute(
                """SELECT * FROM p8.get_entities(%s, %s)""", data=(keys, userid)
            )
            if keys
            else None
        )

        if not data:
            return [
                {
                    "status": "NO DATA",
                    "message": f"There were no data when we fetched {keys=} Please use another method to answer the question or return to the user with a new suggested plan or summary of what you know so far. If you still have different functions to use please try those before completion.",
                }
            ]
        return data[0]

    def search(self, question: str, user_id: str = None):
        """
        If the repository has been activated with a model we use the models search function
        Otherwise we use percolates generic plan and search.
        Either way, feel free to ask a detailed question and we will seek data.

        Args:
            question: detailed natural language question
        """

        """in future we should pardo multiple questions"""
        if isinstance(question, list):
            question = "\n".join(question)

        Q = f"""select * from p8.query_entity(%s,%s, %s) """

        result = self.execute(
            Q, data=(question, self.model.get_model_full_name(), user_id)
        )

        try:
            if result:
                a = result[0].get("relational_result")
                b = result[0].get("vector_result")
                if a is None and b is None:
                    logger.warning(
                        f"Nothing was recovered from the relational or vector result - injecting a prompt"
                    )
                    return [
                        {
                            "status": "no data",
                            "next-steps": f"There is no data to address the questions {question} further using this query but dont worry - try asking for help to find another tool if we have not already been able to find data for this question elsewhere",
                        }
                    ]
        except:
            pass

        return result

    def get_model_database_schema(self):
        assert (
            self.model is not None
        ), "The model is empty - you should construct an instance of the postgres service as a repository(Model)"
        q = f"""SELECT 
            column_name AS field_name,
            data_type AS field_type
            FROM information_schema.columns
            WHERE table_name = %s AND table_schema = %s
        """
        return self.execute(
            q, data=(self.model.get_model_name(), self.model.get_model_namespace())
        )

    def model_registration_script(self, primary: bool = True, secondary: bool = True):
        """for bootstrapping we can generate the model registration script including data
        normally we need to create all the tables first and then we can add aux stuff
        """
        primary_scripts = {
            "register entity": self.helper.create_script(if_not_exists=True),
        }
        secondary_scripts = {
            "register_embeddings": self.helper.create_embedding_table_script(),
            "insert_field_data": SqlModelHelper(ModelField).get_data_load_statement(
                self.helper.get_model_field_models()
            ),
            "insert_agent_data": SqlModelHelper(Agent).get_data_load_statement(
                self.helper.get_model_agent_record()
            ),
            "register_entities": self.helper.get_register_entities_query(),
        }

        statement = ""

        def add_scripts(scripts, s):
            for k, v in scripts.items():
                if v:
                    s += f"\n-- {k} ({self.helper.model.get_model_full_name()})------"
                    s += "\n-- ------------------\n"
                    s += v
                    s += "\n-- ------------------\n"
            return s

        if primary:
            statement = add_scripts(primary_scripts, statement)
        if secondary:
            statement = add_scripts(secondary_scripts, statement)

        return statement

    def register(
        self,
        plan: bool = False,
        register_entities: bool = True,
        make_discoverable: bool = False,
        allow_create_schema: bool = False,
    ):
        """register the entity in percolate
        -- create the type's table and embeddings table
        -- add the fields for the model
        -- add the agent model entity
        -- register the entity which means adding some supporting views etc.
        """
        assert (
            self.model is not None
        ), "You need to specify a model in the constructor or via a repository to register models"
        script = self.helper.create_script()
        # logger.debug(script)
        if plan == True:
            logger.debug(f"Exiting as {plan=}")
            return script

        try:
            self.execute(script, verbose_errors=False)
            logger.debug(f"Created table {self.helper.model.get_model_table_name()}")
        except DuplicateTable:
            logger.warning(
                f"The table already exists - will check for schema migration or ignore"
            )
            current_fields = self.get_model_database_schema()
            script = self.helper.try_generate_migration_script(current_fields)
            if script:
                logger.warning(f"Migrating schema with {script}")
                self.execute(script)

        """added the embedding but check if there are certainly embedding columns"""
        if self.helper.table_has_embeddings:
            try:
                script = self.helper.create_embedding_table_script()
                self.execute(script, verbose_errors=False)
                logger.debug(
                    f"Created embedding table - {self.helper.model.get_model_embedding_table_name()}"
                )
            except DuplicateTable:
                logger.warning(f"The embedding-associated table already exists")

        if make_discoverable:
            """discoverable entities are agents that can be run as functions"""
            self.repository(Function).update_records(Function.from_entity(self.model))

        if register_entities:
            logger.debug("Updating model fields")
            self.repository(ModelField).update_records(
                self.helper.get_model_field_models()
            )

            logger.debug(f"Adding the model agent")
            self.repository(Agent).update_records(self.helper.get_model_agent_record())

            """the registration"""
            if "name" in self.helper.model.model_fields:
                # TODO: we need some way to map a key field for the graph. at the moment a name property is at least implicitly required. We would need this or a business key attribute in the database
                self.execute(
                    "select * from p8.register_entities(%s)",
                    data=(self.helper.model.get_model_full_name(),),
                )

                logger.info(f"Entity registered")
        else:
            logger.info("Done - register entities was disabled")

    def eval_function_call(self, name: str, arguments: dict):
        """call the function which can be rest or native via the database"""
        args = {"function": {"name": name, "arguments": arguments}}
        return self.execute(
            f""" select * from p8.eval_function_call(%s) """, data=(json.dumps(args),)
        )

    def get_tools_metadata(self, names: typing.List[str], scheme: str = "openai"):
        """list tool metadata in the scheme by tool names"""
        data = self.execute(
            f""" select * from p8.get_tools_by_name(%s,%s) """, data=(names, scheme)
        )
        if data:
            return data[0]["get_tools_by_name"]
        return []

    def execute(
        cls,
        query: str,
        data: tuple = None,
        as_upsert: bool = False,
        page_size: int = 100,
        verbose_errors: bool = True,
        timeout_seconds: int = None,
    ):
        """run any sql query - this works only for selects and transactional updates without selects

        Args:
            query:
            data: tuple of args
            as_upsert: hint to use the upsert mode
            page_size: for upsert batching
            verbose_errors:
            timeout_seconds: keep this off but for testing multi turn you can set it to something but keep in mind we would use background workers for this
        """

        if cls.conn is None:
            cls._reopen_connection()
        if not query:
            return

        if timeout_seconds and isinstance(timeout_seconds, int):
            query = f"set statement_timeout = '{timeout_seconds}s';{query}"
        try:
            """we can reopen the connection if needed"""
            try:
                c = cls.conn.cursor()
            except:
                cls._reopen_connection()
                c = cls.conn.cursor()

            """prepare the query"""
            if as_upsert:
                psycopg2.extras.execute_values(
                    c, query, data, template=None, page_size=page_size
                )
            else:
                c.execute(query, data)

            if c.description:
                result = c.fetchall()
                """if we have and updated and read we can commit and send,
                otherwise we commit outside this block"""
                cls.conn.commit()
                column_names = [desc[0] for desc in c.description or []]
                result = [dict(zip(column_names, r)) for r in result]
                return result
            """case of upsert no-query transactions"""
            cls.conn.commit()
        except Exception as pex:
            msg = f"Failing to execute query {query} for model {cls.model} - Postgres error: {pex}, {data}"
            if not verbose_errors:
                msg = f"Failing to execute query model {cls.model} - {verbose_errors=} - {pex}"
            logger.warning(msg)
            cls.conn.rollback()
            raise
        finally:
            cls.conn.close()
            cls.conn = None

    def select(self, fields: typing.List[str] = None, **kwargs):
        """
        select based on the model and use kwargs as quality or in-list template predicates
        """
        assert (
            self.model is not None
        ), "You need to specify a model in the constructor or via a repository to select models"

        data = None
        if kwargs:
            data = tuple(kwargs.values())
        return self.execute(self.helper.select_query(fields, **kwargs), data=data)

    def get_by_name(cls, name: str, as_model: bool = False):
        """select model by name"""
        data = cls.select(name=name)
        if data and as_model and cls.model:
            return [cls.model(**d) for d in data]
        return data

    def get_by_id(cls, id: str, as_model: bool = False) -> dict | AbstractModel:
        """select dictionary values by if unless as model set set - returns one value"""
        data = cls.select(id=id)
        if data:
            data = data[0]
        if as_model and cls.model:
            data = cls.model(**data)
        return data

    def select_to_model(self, fields: typing.List[str] = None):
        """
        like select except we construct the model objects
        """
        return [self.model.model_parse(d) for d in self.select(fields)]

    def execute_upsert(cls, query: str, data: tuple = None, page_size: int = 100):
        """run an upsert sql query"""
        return cls.execute(query, data=data, page_size=page_size, as_upsert=True)

    def update_records(
        self,
        records: typing.List[BaseModel],
        batch_size: int = 1000,
        index_entities: bool = False,
    ):
        """records are updated using typed object relational mapping."""

        if records and not isinstance(records, list):
            records = [records]

        if self.model is None:
            """we encourage explicitly construct repository but we will infer"""
            return self.repository(records[0]).update_records(
                records=records, batch_size=batch_size
            )

        if len(records) > batch_size:
            logger.info(f"Saving  {len(records)} records in batches of {batch_size}")
            for batch in batch_collection(records, batch_size=batch_size):
                sample = self.update_records(
                    batch, batch_size=batch_size, index_entities=index_entities
                )
            return sample

        data = [
            tuple(self.helper.serialize_for_db(r).values())
            for i, r in enumerate(records)
        ]

        if records:
            query = self.helper.upsert_query(batch_size=len(records))
            try:
                result = self.execute_upsert(query=query, data=data)
            except:
                logger.warning(f"Failing to run {query}")
                raise

            if index_entities:
                self.index_entities()
            return result
        else:
            logger.warning(f"Nothing to do - records is empty {records}")

    def index_entity_by_name(self, entity_name: str, id: uuid.UUID = None):
        """
        index entities - a session id can be passed in for the audit callback
        this is very much WIP - it may be this moves into background workers in the database
        """

        assert (
            self.model is not None
        ), "The model is null - did you mean to create a repository with the model first?"

        r1, r2 = None, None
        errors = ""
        try:

            r1 = self.execute(
                f""" select * from p8.insert_entity_nodes(%s); """, data=(entity_name,)
            )
        except Exception as ex:
            errors += traceback.format_exc()
            logger.warning(f"Failed to compute nodes {traceback.format_exc()}")
        try:
            r2 = self.execute(
                f""" select * from p8.insert_entity_embeddings(%s); """,
                data=(entity_name,),
            )
        except Exception as ex:
            errors += traceback.format_exc()
            logger.warning(f"Failed to compute embeddings {ex}")

        metrics = {
            "entities added": r1,
            "embeddings added": r2,
        }

        if not id:
            id = uuid.uuid1()

        if errors == "":
            self.repository(IndexAudit).update_records(
                IndexAudit(
                    id=id,
                    model_name="percolate",
                    entity_full_name=entity_name,
                    metrics=metrics,
                    status="OK",
                    message="Index updated without errors",
                )
            )
        else:
            self.repository(IndexAudit).update_records(
                IndexAudit(
                    id=id,
                    model_name="percolate",
                    entity_full_name=entity_name,
                    metrics=metrics,
                    status="ERROR",
                    message=errors,
                )
            )

        logger.info(metrics)
        return metrics

    def index_entities(self):
        """This is to allow push index but we typically use the DB background workers to do this for us"""

        logger.info(f"indexing entity {self.model}")
        return self.index_entity_by_name(self.model.get_model_full_name())

    def run(
        self, question: str, model: str, max_iterations: int = 5, agent: str = None
    ):
        """this will run the database run function for convenience testing.
        Run executes an agentic loop using the provided agent class.
        Sessions and AIResponses are saved in the database.

        Args:
            question: ask a question
            model: the language model will default to gpt-40
            max_iterations: the number of turns the agent can take - default 5
            agent: the name of the registered agent to use - defaults to the p8.PercolateAgent
        """

        from percolate.utils.env import DEFAULT_MODEL

        agent = agent or self.model.get_model_full_name() if self.model else None
        """default agent"""
        agent = agent or "p8.PercolateAgent"
        model = model or DEFAULT_MODEL
        return self.execute(
            f"""SELECT * FROM run(%s,%s, %s, %s)""",
            data=(question, max_iterations, model, agent),
        )



"""
Models for audio processing in the Percolate framework.
These models represent audio files, chunks, and processing statuses for the audio pipeline.
"""

from pydantic import BaseModel, Field, model_validator
from typing import Optional, List, Dict, Any, Union
from datetime import datetime, timezone
import uuid
from ..AbstractModel import AbstractModel
from percolate.models import DefaultEmbeddingField

class AudioProcessingStatus:
    """Status of an audio file in the processing pipeline"""
    UPLOADING = "uploading"
    UPLOADED = "uploaded"
    PROCESSING = "processing"
    CHUNKING = "chunking"
    TRANSCRIBING = "transcribing"
    COMPLETED = "completed"
    FAILED = "failed"

class AudioFile(AbstractModel):
    """Model representing an uploaded audio file"""
    model_config = {'namespace': 'public'}
    
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    userid: Optional[str | uuid.UUID] = Field(default=None, description="The user id if known")
    project_name: str
    filename: str
    file_size: int
    content_type: str
    duration: Optional[float] = None
    upload_date: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    status: str = AudioProcessingStatus.UPLOADING
    s3_uri: str
    chunks: Optional[List["AudioChunk"]] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    
    @model_validator(mode='before')
    @classmethod
    def _v(cls, values):
        """Generate ID if not provided or convert string to UUID"""
        if 'id' not in values or not values['id']:
            values['id'] = uuid.uuid4()
        elif isinstance(values['id'], str):
            # Convert string to UUID if needed
            values['id'] = uuid.UUID(values['id'])
        return values

class AudioChunk(AbstractModel):
    """Model representing a chunk of an audio file for processing"""
    model_config = {'namespace': 'public'}
    
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    audio_file_id: uuid.UUID
    start_time: float
    end_time: float
    duration: float
    s3_uri: str
    transcription: Optional[str] = DefaultEmbeddingField(default='', description='transcribed audio is a resource')
    confidence: Optional[float] = None
    speaker_id: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    userid: Optional[str|uuid.UUID] = Field(default=None, description="the audio chunk belongs to a user")
    
    @model_validator(mode='before')
    @classmethod
    def _v(cls, values):
        """Generate ID if not provided or convert string to UUID"""
        if 'id' not in values or not values['id']:
            values['id'] = uuid.uuid4()
        elif isinstance(values['id'], str):
            # Convert string to UUID if needed
            values['id'] = uuid.UUID(values['id'])
            
        # Convert audio_file_id to UUID if it's a string
        if 'audio_file_id' in values and isinstance(values['audio_file_id'], str):
            values['audio_file_id'] = uuid.UUID(values['audio_file_id'])
            
        return values

class AudioUploadResponse(BaseModel):
    """Response model for audio upload endpoint"""
    file_id: uuid.UUID
    filename: str
    status: str
    s3_uri: str

class ChunkingConfig(BaseModel):
    """Configuration for audio chunking"""
    max_chunk_size: float = Field(default=30.0, description="Maximum chunk size in seconds")
    min_chunk_size: float = Field(default=1.0, description="Minimum chunk size in seconds")
    merge_threshold: float = Field(default=0.5, description="Silence threshold for merging in seconds")
    vad_threshold: float = Field(default=0.5, description="Voice activity detection threshold (0.0-1.0)")

class TranscriptionConfig(BaseModel):
    """Configuration for transcription"""
    model: str = Field(default="whisper-1", description="OpenAI Whisper model to use")
    language: Optional[str] = None
    temperature: float = Field(default=0.0, description="Sampling temperature (0.0-1.0)")
    prompt: Optional[str] = None

class AudioPipelineConfig(BaseModel):
    """Configuration for the audio pipeline"""
    chunking: ChunkingConfig = Field(default_factory=ChunkingConfig)
    transcription: TranscriptionConfig = Field(default_factory=TranscriptionConfig)
    retention_period: int = Field(default=90, description="Days to retain audio files in S3")

class AudioPipeline(AbstractModel):
    """Model for tracking audio pipeline execution"""
    model_config = {'namespace': 'public'}
    
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    audio_file_id: str|uuid.UUID
    status: str
    started_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    completed_at: Optional[datetime] = None
    config: Dict[str, Any] = Field(default_factory=dict)
    error_message: Optional[str] = None
    
    @model_validator(mode='before')
    @classmethod
    def _v(cls, values):
        """Generate ID if not provided or convert string to UUID"""
        if 'id' not in values or not values['id']:
            values['id'] = uuid.uuid4()
        elif isinstance(values['id'], str):
            # Convert string to UUID if needed
            values['id'] = uuid.UUID(values['id'])
            
        # Convert audio_file_id to UUID if it's a string
        if 'audio_file_id' in values and isinstance(values['audio_file_id'], str):
            values['audio_file_id'] = uuid.UUID(values['audio_file_id'])
            
        # Set default status if not provided
        if 'status' not in values:
            values['status'] = AudioProcessingStatus.PROCESSING
            
        return values

class AudioResource(AbstractModel):
    """Model for storing audio resources in Percolate"""
    model_config = {'namespace': 'public'}
    
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    uri: str
    name: str
    content_type: str
    content: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    
    @model_validator(mode='before')
    @classmethod
    def _v(cls, values):
        """Generate ID if not provided or convert string to UUID"""
        if 'id' not in values or not values['id']:
            values['id'] = uuid.uuid4()
        elif isinstance(values['id'], str):
            # Convert string to UUID if needed
            values['id'] = uuid.UUID(values['id'])
        return values

def register_audio_models():
    """
    Register all audio models with the Percolate database.
    This function should be called during application startup to ensure 
    all audio-related models are properly registered with the database.
    
    Returns:
        Dict with registration results
    """
    import percolate as p8
    
    models = [
        AudioFile,
        AudioChunk,
        AudioPipeline,
        AudioResource
    ]
    
    results = {}
    
    for model in models:
        model_name = model.__name__
        try:
            # Register the model with Percolate
            p8.repository(model).register()
            results[model_name] = "Registered successfully"
        except Exception as e:
            results[model_name] = f"Failed: {str(e)}"
    
    return results

# Reference to support circular references
AudioFile.model_rebuild()


from pydantic import Field, model_validator
import typing
from ..AbstractModel import AbstractModel 
import uuid

def create_lookup(
    data: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    """in a DAG where we are being efficient with named refs, we can get a lookup of all nodes with this and then expand for pydantic"""
    lookup = {}

    def _traverse(node: typing.Dict[str, typing.Any]):
        if "plan_description" in node:
            if node["name"] not in lookup:
                lookup[node["name"]] = {}
            lookup[node["name"]].update(node)
        if "depends" in node:
            for dep in node["depends"]:
                _traverse(dep)

    _traverse(data)
    return lookup


def expand_refs(
    node: typing.Dict[str, typing.Any],
    lookup: typing.Dict[str, typing.Dict[str, typing.Any]],
) -> typing.Dict[str, typing.Any]:
    """in a DAG where we are being efficient with named refs, we can get a lookup of all nodes with this and then expand for pydantic using this function"""
    if "depends" in node:
        expanded_depends = []
        for dep in node["depends"]:
            dep_name = dep["name"]
            if dep_name in lookup:
                expanded_dep = expand_refs(lookup[dep_name], lookup)
                expanded_depends.append(expanded_dep)
            else:
                expanded_depends.append(dep)
        node["depends"] = expanded_depends
    return node


class PlanFunctions(AbstractModel):
    name: str = Field(
        description="fully qualified function name e.g. <namespace>_<name>"
    )
    description: str = Field(
        description="a description of the function preferably with the current context taken into account e.g. provide good example parameters"
    )
    rating: float = Field(
        description="a rating from 0 to 100 for how useful this function should be in context"
    )


class PlanModel(AbstractModel):
    """
    You are an agent that plans function calling and agent resources for other agents but you do not call functions yourself.
    Respond to the caller only with the context you ascertain from the data you are given.
    You should construct a Directed Acyclic Graph (RAG) of one or more functions that can be called.
    Observe where there are dependencies between functions
    """

    id : typing.Optional[uuid.UUID | str ] = Field(None, description="An id for the plan - better to set this but we can add them uniquely to users and sessions to")
    name: typing.Optional[str] = Field(
        description="The unique name of the plan node", default=None
    )

    plan_description: str = Field(
        description="The plan to prompt the agent - should provide fully strategy and explain what dependencies exist with other stages"
    )
    questions: typing.Optional[typing.List[str]] = Field(
        description="The question in this plan instance as the user would ask it. A plan can be constructed without a clear question",
        default=None,
    )
    extra_arguments: typing.Optional[dict] = Field(
        description="Placeholder/hint for extra parameters that should be passed from previous stages such as data or identifiers that were discovered in the data and expected by the function either as a parameter or important context",
        default=None,
    )
    functions: typing.Optional[typing.List[PlanFunctions]] = Field(
        description="A collection of functions designed for use with this context",
        default=None,
    )
    depends: typing.Optional[typing.List["PlanModel"]] = Field(
        description="A dependency graph - plans can be chained into waves of functions that can be called in parallel or one after the other. Data dependencies are injected to downstream plans",
        default=None,
    )
    userid: typing.Optional[str] = Field(None, description="A user that owns the plan")

    @model_validator(mode="before")
    @classmethod
    def _expand(cls, values):
        """expand entity refs for pydantic model"""
        l = create_lookup(values)
        values = expand_refs(values, l)
        return values

    # @property
    # def plan_dependency_image(cls) -> Image.Image:
    #     """A PNG format of the plan"""
    #     return Image.open(BytesIO(cls.draw_plan_dependency_graph()._repr_image_png()))

    def draw_plan_dependency_graph(cls):
        """simple plan viz"""
        from graphviz import Digraph

        dot = Digraph(comment="Plan Dependency Graph")

        def add_node(node, parent=None):
            node_label = f'{node["name"]}\n{node["plan_description"]}'
            dot.node(node["name"], label=node_label, shape="box")

            if parent:
                dot.edge(node["name"], parent, dir="back")

            for func in node["functions"]:
                func_label = f'{func["name"]}({", ".join(func.get("arguments", []))})'
                dot.node(func_label, shape="ellipse")
                dot.edge(node["name"], func_label)

            if node["depends"]:
                for dependency in node["depends"]:
                    add_node(dependency, node["name"])

        add_node(cls.model_dump())

        return dot
